from mesido.pycml import Connector, Variable

from ._internal import QTHComponent


class QTHPort(QTHComponent, Connector):
    """
    Connector with potential temperature (T), flow discharge (Q) and head (H)
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(Variable, "T")  # Temperature
        self.add_variable(Variable, "Q")  # Volume flow (positive inwards)
        self.add_variable(Variable, "H")  # Head



from mesido.pycml import Connector, Variable
from mesido.pycml.component_library.milp._internal import HeatComponent


class HeatPort(HeatComponent, Connector):
    """
    The HeatPort is used to model the variables at an in or outgoing port of a component. For the
    HeatMixin we model thermal Power (Heat [W]), flow (Q [m3/s]) and head (H [m]) at every port in
    the network.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(Variable, "Heat")
        self.add_variable(Variable, "Q")
        self.add_variable(Variable, "H")
        self.add_variable(Variable, "Hydraulic_power")



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import (
    demand_matching_test,
    electric_power_conservation_test,
    energy_conservation_test,
    heat_to_discharge_test,
)


class TestElecBoiler(TestCase):
    def test_elec_boiler(self):
        """
        This tests checks the elec boiler for the standard checks and the energy conservation over
        the commodity change.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the heat and electricity commodity
        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_witheboiler.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)
        electric_power_conservation_test(heat_problem, results)

        np.testing.assert_array_less(0.0, results["ElectricBoiler_9aab.Heat_source"])
        np.testing.assert_array_less(0.0, results["ElectricityProducer_4dde.ElectricityOut.Power"])
        np.testing.assert_array_less(
            parameters["ElectricBoiler_9aab.efficiency"]
            * results["ElectricBoiler_9aab.Power_consumed"],
            results["ElectricBoiler_9aab.Heat_source"] + 1.0e-6,
        )

    def test_air_water_hp_elec(self):
        """
        This tests checks the air-water hp elec for the standard checks and the energy conservation
        over the commodity change.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the commodity
        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_withHP.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)
        electric_power_conservation_test(heat_problem, results)

        np.testing.assert_array_less(0.0, results["HeatPump_d8fd.Heat_source"])
        np.testing.assert_array_less(0.0, results["ElectricityProducer_4dde.ElectricityOut.Power"])
        np.testing.assert_array_less(
            parameters["HeatPump_d8fd.cop"] * results["HeatPump_d8fd.Power_consumed"],
            results["HeatPump_d8fd.Heat_source"] + 1.0e-6,
        )



from strenum import StrEnum


class NetworkSettings(StrEnum):
    """
    Enumeration for network settings.
    """

    NETWORK_TYPE_GAS = "Gas"  # Natural gas
    NETWORK_TYPE_HYDROGEN = "Hydrogen"
    NETWORK_TYPE_HEAT = "Heat"
    NETWORK_TYPE_ELECTRICITY = "Electricity"

    NETWORK_COMPOSITION_GAS = (
        "methane[0.813]&ethane[0.0285]&propane[0.0037]&butane[0.0014]"
        "&n-Pentane[0.0004]&n-Hexane[0.0005]&nitrogen[0.1435]&CO2[0.0089]"
        "&oxygen[0.0001]"
    )  # Natural gas Groningen



from numpy import nan

from .electricity_base import ElectricityTwoPort
from .._internal import BaseAsset


class Transformer(ElectricityTwoPort, BaseAsset):
    """
    The transformer component is used to change the voltage level of the network whilst conserving
    the power.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "transformer"
        self.meret_place = 1

        self.price = nan

        self.min_voltage = nan

        self.power_nominal = nan

        self.add_equation(
            ((self.ElectricityIn.Power - self.ElectricityOut.Power) / self.power_nominal)
        )



from .airwater_heat_pump_elec import AirWaterHeatPumpElec
from .electro_boiler import ElecBoiler
from .electrolyzer import Electrolyzer
from .gas_boiler import GasBoiler

__all__ = [
    "AirWaterHeatPumpElec",
    "ElecBoiler",
    "Electrolyzer",
    "GasBoiler",
]



import json
import logging
import math
import os
from pathlib import Path
from typing import Any, Dict, Tuple, Type, Union

import esdl
from esdl import TimeUnitEnum, UnitEnum

from mesido.esdl._exceptions import _RetryLaterException
from mesido.esdl.common import Asset
from mesido.pycml import Model as _Model


logger = logging.getLogger("mesido")

MODIFIERS = Dict[str, Union[str, int, float]]

HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS = 4200 * 988
WATTHOUR_TO_JOULE = 3600

MULTI_ENUM_NAME_TO_FACTOR = {
    esdl.MultiplierEnum.ATTO: 1e-18,
    esdl.MultiplierEnum.FEMTO: 1e-15,
    esdl.MultiplierEnum.PICO: 1e-12,
    esdl.MultiplierEnum.NANO: 1e-9,
    esdl.MultiplierEnum.MICRO: 1e-6,
    esdl.MultiplierEnum.MILLI: 1e-3,
    esdl.MultiplierEnum.CENTI: 1e-2,
    esdl.MultiplierEnum.DECI: 1e-1,
    esdl.MultiplierEnum.NONE: 1e0,
    esdl.MultiplierEnum.DEKA: 1e1,
    esdl.MultiplierEnum.HECTO: 1e2,
    esdl.MultiplierEnum.KILO: 1e3,
    esdl.MultiplierEnum.MEGA: 1e6,
    esdl.MultiplierEnum.GIGA: 1e9,
    esdl.MultiplierEnum.TERA: 1e12,
    esdl.MultiplierEnum.TERRA: 1e12,
    esdl.MultiplierEnum.PETA: 1e15,
    esdl.MultiplierEnum.EXA: 1e18,
}


class _AssetToComponentBase:
    # A map of pipe class name to edr asset in _edr_pipes.json
    STEEL_S1_PIPE_EDR_ASSETS = {
        "DN20": "Steel-S1-DN-20",
        "DN25": "Steel-S1-DN-25",
        "DN32": "Steel-S1-DN-32",
        "DN40": "Steel-S1-DN-40",
        "DN50": "Steel-S1-DN-50",
        "DN65": "Steel-S1-DN-65",
        "DN80": "Steel-S1-DN-80",
        "DN100": "Steel-S1-DN-100",
        "DN125": "Steel-S1-DN-125",
        "DN150": "Steel-S1-DN-150",
        "DN200": "Steel-S1-DN-200",
        "DN250": "Steel-S1-DN-250",
        "DN300": "Steel-S1-DN-300",
        "DN350": "Steel-S1-DN-350",
        "DN400": "Steel-S1-DN-400",
        "DN450": "Steel-S1-DN-450",
        "DN500": "Steel-S1-DN-500",
        "DN600": "Steel-S1-DN-600",
        "DN700": "Steel-S1-DN-700",
        "DN800": "Steel-S1-DN-800",
        "DN900": "Steel-S1-DN-900",
        "DN1000": "Steel-S1-DN-1000",
        "DN1100": "Steel-S1-DN-1100",
        "DN1200": "Steel-S1-DN-1200",
    }
    # A map of the esdl assets to the asset types in pycml
    component_map = {
        "Airco": "airco",
        "ATES": "ates",
        "Battery": "electricity_storage",
        "Bus": "electricity_node",
        "ElectricBoiler": "elec_boiler",
        "ElectricityCable": "electricity_cable",
        "ElectricityDemand": "electricity_demand",
        "ElectricityProducer": "electricity_source",
        "Electrolyzer": "electrolyzer",
        "Export": "export",
        "Compressor": "compressor",
        "GenericConsumer": "heat_demand",
        "CoolingDemand": "cold_demand",
        "HeatExchange": "heat_exchanger",
        "HeatingDemand": "heat_demand",
        "HeatPump": "heat_pump",
        "GasHeater": "gas_boiler",
        "GasProducer": "gas_source",
        "GasDemand": "gas_demand",
        "GasConversion": "gas_substation",
        "GasStorage": "gas_tank_storage",
        "GenericProducer": "heat_source",
        "GeothermalSource": "heat_source",
        "Losses": "heat_demand",
        "HeatProducer": "heat_source",
        "Import": "import",
        "ResidualHeatSource": "heat_source",
        "GenericConversion": "heat_exchanger",
        "Joint": "node",
        "Pipe": "heat_pipe",
        "Pump": "pump",
        "PressureReducingValve": "gas_substation",
        "PVInstallation": "electricity_source",
        "HeatStorage": "heat_buffer",
        "Sensor": "skip",
        "Valve": "control_valve",
        "WindPark": "electricity_source",
        "Transformer": "transformer",
        "CheckValve": "check_valve",
    }

    primary_port_name_convention = "primary"
    secondary_port_name_convention = "secondary"

    def __init__(self, **kwargs):
        """
        In this init we initialize some dicts and we load the edr pipes.
        """
        self._port_to_q_nominal = {}
        self._port_to_q_max = {}
        self._port_to_i_nominal = {}
        self._port_to_i_max = {}
        self._port_to_esdl_component_type = {}
        self._edr_pipes = json.load(
            open(os.path.join(Path(__file__).parent, "_edr_pipes.json"), "r")
        )

    def convert(self, asset: Asset) -> Tuple[Type[_Model], MODIFIERS]:
        """
        Converts an asset to a PyCML Heat component type and its modifiers.

        With more descriptive variable names the return type would be:
            Tuple[pycml_heat_component_type, Dict[component_attribute, new_attribute_value]]
        """
        ports = []
        if asset.in_ports is not None:
            ports.extend(asset.in_ports)
        if asset.out_ports is not None:
            ports.extend(asset.out_ports)
        assert len(ports) > 0

        for port in ports:
            self._port_to_esdl_component_type[port] = asset.asset_type

        dispatch_method_name = f"convert_{self.component_map[asset.asset_type]}"
        return getattr(self, dispatch_method_name)(asset)

    def _pipe_get_diameter_and_insulation(self, asset: Asset) -> Tuple[float, list, list]:
        """
        There are multiple ways to specify pipe properties like inner-diameter and
        pipe/insulation material and thickness.  The user specified nominal diameter (DN size)
        takes precedence over potential user specified innerDiameter and material (while logging
        warnings when either of these two variables are specified in combination with the pipe DN)
        Parameters
        ----------
        asset : Asset pipe object with it's properties from ESDL

        Returns
        -------
        pipe inner diameter, thickness and conductivity of each insulation layer
        """

        full_name = f"{asset.asset_type} '{asset.name}'"
        if asset.attributes["innerDiameter"] and asset.attributes["diameter"].value > 0:
            logger.warning(
                f"{full_name}' has both 'innerDiameter' and 'diameter' specified. "
                f"Diameter of {asset.attributes['diameter'].name} will be used."
            )
        if asset.attributes["material"] and asset.attributes["diameter"].value > 0:
            logger.warning(
                f"{full_name}' has both 'material' and 'diameter' specified. "
                f"Insulation properties of {asset.attributes['diameter'].name} will be used."
            )
        if asset.attributes["material"] and (
            asset.attributes["diameter"].value == 0 and not asset.attributes["innerDiameter"]
        ):
            logger.warning(
                f"{full_name}' has only 'material' specified, but no information on diameter. "
                f"Diameter and insulation properties of DN200 will be used."
            )
        if asset.attributes["diameter"].value == 0 and not asset.attributes["innerDiameter"]:
            if asset.attributes["material"]:
                logger.warning(
                    f"{full_name}' has only 'material' specified, but no information on diameter. "
                    f"Diameter and insulation properties of DN200 will be used."
                )
            else:
                logger.warning(
                    f"{full_name}' has no DN size or innerDiameter specified. "
                    f"Diameter and insulation properties of DN200 will be used. "
                )

        edr_dn_size = None
        if asset.attributes["diameter"].value > 0:
            edr_dn_size = str(asset.attributes["diameter"].name)
        elif not asset.attributes["innerDiameter"]:
            edr_dn_size = "DN200"

        # NaN means the default values will be used
        insulation_thicknesses = math.nan
        conductivies_insulation = math.nan

        if edr_dn_size:
            # Get insulation and diameter properties from EDR asset with this size.
            edr_asset = self._edr_pipes[self.STEEL_S1_PIPE_EDR_ASSETS[edr_dn_size]]
            inner_diameter = edr_asset["inner_diameter"]
            insulation_thicknesses = edr_asset["insulation_thicknesses"]
            conductivies_insulation = edr_asset["conductivies_insulation"]
        else:
            assert asset.attributes["innerDiameter"]
            inner_diameter = asset.attributes["innerDiameter"]

            # Insulation properties
            material = asset.attributes["material"]

            if material is not None:
                if isinstance(material, esdl.esdl.MatterReference):
                    material = material.reference

                assert isinstance(material, esdl.esdl.CompoundMatter)
                components = material.component.items
                if components:
                    insulation_thicknesses = [x.layerWidth for x in components]
                    conductivies_insulation = [x.matter.thermalConductivity for x in components]

        return inner_diameter, insulation_thicknesses, conductivies_insulation

    def _is_disconnectable_pipe(self, asset: Asset) -> bool:
        """
        This function checks if the pipe is connected to specific assets (e.g. source) and if so
        returns true. The true here means that we will later make a is_disconnected variable
        allowing for optionally disconnecting a pipe from the optimization meaning it will not have
        any flow, but also avoiding the need to compensate the milp losses for that pipe.

        Parameters
        ----------
        asset : The asset object of an pipe

        Returns
        -------
        A bool that specifies whether we should have a disconnectable variable for this
        pipe.
        """
        assert asset.asset_type == "Pipe"
        if len(asset.in_ports) == 1 and len(asset.out_ports) == 1:
            connected_type_in = self._port_to_esdl_component_type.get(
                asset.in_ports[0].connectedTo[0], None
            )
            connected_type_out = self._port_to_esdl_component_type.get(
                asset.out_ports[0].connectedTo[0], None
            )
        else:
            raise RuntimeError("Pipe does not have 1 in port and 1 out port")
        # TODO: add other components which can be disabled and thus of which the pipes are allowed
        #  to be disabled: , "heat_exchanger", "heat_pump", "ates"
        types = {
            k
            for k, v in self.component_map.items()
            if v
            in {
                "heat_source",
                "heat_buffer",
                "ates",
                "heat_exchanger",
                "heat_pump",
                "cold_demand",
            }
        }

        if types.intersection({connected_type_in, connected_type_out}):
            return True
        elif connected_type_in is None or connected_type_out is None:
            raise _RetryLaterException(
                f"Could not determine if {asset.asset_type} '{asset.name}' "
                f"is a source or buffer pipe"
            )
        else:
            return False

    def _set_q_nominal(self, asset: Asset, q_nominal: float) -> None:
        """
        This function populates a dict with the nominal volumetric flow in m3/s for the ports of all
        pipes.

        Parameters
        ----------
        asset :
        q_nominal : float of the nominal flow through that pipe

        Returns
        -------
        None
        """
        try:
            self._port_to_q_nominal[asset.in_ports[0]] = q_nominal
        except TypeError:
            pass
        try:
            self._port_to_q_nominal[asset.out_ports[0]] = q_nominal
        except TypeError:
            pass

    def _set_q_max(self, asset: Asset, q_max: float) -> None:
        """
        This function populates a dict with the max volumetric flow in m3/s for the ports of all
        pipes.

        Parameters
        ----------
        asset :
        q_max : float of the max flow through that pipe

        Returns
        -------
        None
        """
        try:
            self._port_to_q_max[asset.in_ports[0]] = q_max
        except TypeError:
            pass
        try:
            self._port_to_q_max[asset.out_ports[0]] = q_max
        except TypeError:
            pass

    def _set_electricity_current_nominal_and_max(
        self, asset: Asset, i_nominal: float, i_max: float
    ) -> None:
        """
        This function populates a dict with the electricity current nominals [A] for the ports of
        all electricity cables.

        Parameters
        ----------
        asset :
        q_max : float of the electricity current nominal

        Returns
        -------
        None
        """
        try:
            self._port_to_i_nominal[asset.in_ports[0]] = i_nominal
            self._port_to_i_max[asset.in_ports[0]] = i_max
        except TypeError:
            pass
        try:
            self._port_to_i_nominal[asset.out_ports[0]] = i_nominal
            self._port_to_i_max[asset.out_ports[0]] = i_max
        except TypeError:
            pass

    def _get_connected_q_max(self, asset: Asset) -> float:
        if asset.in_ports is None or asset.asset_type == "Electrolyzer":
            connected_port = asset.out_ports[0].connectedTo[0]
            q_max = self._port_to_q_max.get(connected_port, None)
            if q_max is not None:
                self._set_q_nominal(asset, q_max)
                return q_max
            else:
                raise _RetryLaterException(
                    f"Could not determine max discharge for {asset.asset_type} '{asset.name}'"
                )
        elif asset.out_ports is None:
            connected_port = asset.in_ports[0].connectedTo[0]
            q_max = self._port_to_q_max.get(connected_port, None)
            if q_max is not None:
                self._set_q_max(asset, q_max)
                return q_max
            else:
                raise _RetryLaterException(
                    f"Could not determine max discharge for {asset.asset_type} '{asset.name}'"
                )

    def _get_connected_i_nominal_and_max(self, asset: Asset) -> Tuple[float, float]:
        if (
            asset.in_ports is not None
            and asset.out_ports is not None
            and len(asset.in_ports) == 1
            and len(asset.out_ports) == 1
            and isinstance(asset.in_ports[0].carrier, esdl.ElectricityCommodity)
            and isinstance(asset.out_ports[0].carrier, esdl.ElectricityCommodity)
        ):  # Transformer
            connected_port = asset.out_ports[0].connectedTo[0]
            i_max_out = self._port_to_i_max.get(connected_port, None)
            i_nom_out = self._port_to_i_nominal.get(connected_port, None)
            connected_port = asset.in_ports[0].connectedTo[0]
            i_max_in = self._port_to_i_max.get(connected_port, None)
            i_nom_in = self._port_to_i_nominal.get(connected_port, None)
            if i_nom_in is not None and i_nom_out is not None:
                self._port_to_i_nominal[asset.in_ports[0]] = i_nom_in
                self._port_to_i_max[asset.in_ports[0]] = i_max_in
                self._port_to_i_nominal[asset.out_ports[0]] = i_nom_out
                self._port_to_i_max[asset.out_ports[0]] = i_max_out
                return i_max_in, i_nom_in, i_max_out, i_nom_out
            else:
                raise _RetryLaterException(
                    f"Could not determine max and nominal current for {asset.asset_type}"
                    " '{asset.name}'"
                )

        elif asset.in_ports is None:
            connected_port = asset.out_ports[0].connectedTo[0]
            i_max = self._port_to_i_max.get(connected_port, None)
            i_nom = self._port_to_i_nominal.get(connected_port, None)
            if i_max is not None:
                self._set_electricity_current_nominal_and_max(asset, i_max, i_nom)
                return i_max, i_nom
            else:
                raise _RetryLaterException(
                    f"Could not determine max and nominal current for {asset.asset_type}"
                    " '{asset.name}'"
                )
        elif (
            asset.out_ports is None
            or asset.asset_type == "Electrolyzer"
            or asset.asset_type == "ElectricBoiler"
            or asset.asset_type == "HeatPump"
        ):
            for port in asset.in_ports:
                if isinstance(port.carrier, esdl.ElectricityCommodity):
                    connected_port = port.connectedTo[0]
            i_max = self._port_to_i_max.get(connected_port, None)
            i_nom = self._port_to_i_nominal.get(connected_port, None)
            if i_max is not None:
                self._set_electricity_current_nominal_and_max(asset, i_max, i_nom)
                return i_max, i_nom
            else:
                raise _RetryLaterException(
                    f"Could not determine max and nominal current for {asset.asset_type}"
                    f" '{asset.name}'"
                )

    def _get_connected_q_nominal(self, asset: Asset) -> Union[float, Dict]:
        """
        This function returns the nominal volumetric flow in m3/s for an asset by checking the dict
        that has all q_nominal for the ports of all pipes. Since all ports must have at least one
        pipe connected to them, this allows us to find all needed nominals. Assets can either be
        connected to one or two hydraulic systems.

        Parameters
        ----------
        asset : Asset object used to check to which ports the asset is connected

        Returns
        -------
        Either the connected nominal flow [m3/s] if it is only connected to one hydraulic system,
        otherwise a dict with the flow nominals of both the primary and secondary side.
        """
        if asset.in_ports is None:
            connected_port = asset.out_ports[0].connectedTo[0]
            q_nominal = self._port_to_q_nominal.get(connected_port, None)
            if q_nominal is not None:
                self._set_q_nominal(asset, q_nominal)
                return q_nominal
            else:
                raise _RetryLaterException(
                    f"Could not determine nominal discharge for {asset.asset_type} '{asset.name}'"
                )
        elif asset.out_ports is None:
            connected_port = asset.in_ports[0].connectedTo[0]
            q_nominal = self._port_to_q_nominal.get(connected_port, None)
            if q_nominal is not None:
                self._set_q_nominal(asset, q_nominal)
                return q_nominal
            else:
                raise _RetryLaterException(
                    f"Could not determine nominal discharge for {asset.asset_type} '{asset.name}'"
                )
        elif (
            len(asset.in_ports) == 1
            and len(asset.out_ports) == 1
            and asset.in_ports[0].carrier.id != asset.out_ports[0].carrier.id
            and isinstance(asset.in_ports[0].carrier, esdl.GasCommodity)
            and isinstance(asset.out_ports[0].carrier, esdl.GasCommodity)
        ):  # Cater for gas substation
            try:
                connected_port = asset.in_ports[0].connectedTo[0]
                q_nominal_in = self._port_to_q_nominal[connected_port]
            except KeyError:
                connected_port = asset.out_ports[0].connectedTo[0]
                q_nominal_out = self._port_to_q_nominal.get(connected_port, None)

            try:
                connected_port = asset.out_ports[0].connectedTo[0]
                q_nominal_out = self._port_to_q_nominal[connected_port]
            except KeyError:
                connected_port = asset.in_ports[0].connectedTo[0]
                q_nominal_in = self._port_to_q_nominal.get(connected_port, None)

            if q_nominal_in is not None and q_nominal_out is not None:
                self._port_to_q_nominal[asset.in_ports[0]] = q_nominal_in
                self._port_to_q_nominal[asset.out_ports[0]] = q_nominal_out
                return q_nominal_in, q_nominal_out
            else:
                raise _RetryLaterException(
                    f"Could not determine nominal discharge for {asset.asset_type} '{asset.name}'"
                )
        elif len(asset.in_ports) == 1 and len(asset.out_ports) == 1:
            try:
                connected_port = asset.in_ports[0].connectedTo[0]
                q_nominal = self._port_to_q_nominal[connected_port]
            except KeyError:
                connected_port = asset.out_ports[0].connectedTo[0]
                q_nominal = self._port_to_q_nominal.get(connected_port, None)

            if q_nominal is not None:
                self._set_q_nominal(asset, q_nominal)
                return q_nominal
            else:
                raise _RetryLaterException(
                    f"Could not determine nominal discharge for {asset.asset_type} '{asset.name}'"
                )
        elif len(asset.in_ports) == 2 and len(asset.out_ports) == 1:  # for gas_boiler or e_boiler
            q_nominals = {}
            try:
                for port in asset.in_ports:
                    connected_port = asset.in_ports[0].connectedTo[0]
                    if isinstance(port.carrier, esdl.GasCommodity):
                        q_nominals["Q_nominal_gas"] = self._port_to_q_nominal[connected_port]
                        self._port_to_q_nominal[port] = q_nominals["Q_nominal_gas"]
                    elif isinstance(port.carrier, esdl.HeatCommodity):
                        q_nominals["Q_nominal"] = self._port_to_q_nominal[connected_port]
                        self._port_to_q_nominal[port] = q_nominals["Q_nominal"]
                    else:
                        logger.error(
                            f"{asset.name} should have at least gas or heat specified on"
                            f"one of the in ports"
                        )
            except KeyError:
                if isinstance(asset.out_ports[0].carrier, esdl.GasCommodity):
                    connected_port = asset.out_ports[0].connectedTo[0]
                    q_nominals["Q_nominal"] = self._port_to_q_nominal.get(connected_port, None)
                else:
                    logger.error(f"{asset.name} should have a heat carrier on out port")

            if q_nominals["Q_nominal"] is not None:
                self._port_to_q_nominal[asset.out_ports[0]] = q_nominals["Q_nominal"]
                return q_nominals
            else:
                raise _RetryLaterException(
                    f"Could not determine nominal discharge for {asset.asset_type} '{asset.name}'"
                )
        elif len(asset.in_ports) >= 2 and len(asset.out_ports) == 2:
            q_nominals = {}
            for p in asset.in_ports:
                if isinstance(p.carrier, esdl.HeatCommodity):
                    out_port = None
                    for p2 in asset.out_ports:
                        if p2.carrier.name.replace("_ret", "") == p.carrier.name.replace(
                            "_ret", ""
                        ):
                            out_port = p2
                    try:
                        connected_port = p.connectedTo[0]
                        q_nominal = self._port_to_q_nominal[connected_port]
                    except KeyError:
                        connected_port = out_port.connectedTo[0]
                        q_nominal = self._port_to_q_nominal.get(connected_port, None)
                    if q_nominal is not None:
                        self._port_to_q_nominal[p] = q_nominal
                        self._port_to_q_nominal[out_port] = q_nominal
                        if "_ret" in p.carrier.name:
                            q_nominals["Secondary"] = {"Q_nominal": q_nominal}
                        else:
                            q_nominals["Primary"] = {"Q_nominal": q_nominal}
                    else:
                        raise _RetryLaterException(
                            f"Could not determine nominal discharge for {asset.asset_type} "
                            f"{asset.name}"
                        )
            return q_nominals

    def _get_cost_figure_modifiers(self, asset: Asset) -> Dict:
        """
        This function takes in an asset and creates a dict with the relevant cost information of
        that asset which is used in the optimization. At this moment we have a four element cost
        structure with:
        InvestmentCost: Scales with asset size
        InstallationCost: Scales with the _aggregation count integer (cost for placement),
                            independent of the size of the individual aggregation counts
        FixedOperationalCost: Yearly operational cost that scales with asset size.
        VariableOperationalCost: Yearly operational cost that scales with asset use.

        Parameters
        ----------
        asset : Asset object to retrieve cost information from.

        Returns
        -------
        Dict with the mentioned cost elements
        """
        modifiers = {}

        if asset.attributes["costInformation"] is None:
            RuntimeWarning(f"{asset.name} has no cost information specified")
            return modifiers

        if asset.asset_type == "HeatStorage":
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
            modifiers["fixed_operational_cost_coefficient"] = self.get_fixed_opex_costs(asset)
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.JOULE
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)
        elif asset.asset_type == "Pipe":
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.METRE
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)
        elif asset.asset_type == "HeatingDemand":
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.WATT
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)
        elif asset.asset_type == "GasDemand":
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
        elif asset.asset_type == "Electrolyzer":
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
            modifiers["fixed_operational_cost_coefficient"] = self.get_fixed_opex_costs(asset)
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(asset)
        elif asset.asset_type == "GasStorage":
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
            modifiers["fixed_operational_cost_coefficient"] = self.get_fixed_opex_costs(asset)
        else:
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
            modifiers["fixed_operational_cost_coefficient"] = self.get_fixed_opex_costs(asset)
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.WATT
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)

        return modifiers

    @staticmethod
    def _get_supply_return_temperatures(asset: Asset) -> Tuple[float, float]:
        """
        This function returns the supply and return temperature for an asset that is connected to
        one hydraulic system.

        Parameters
        ----------
        asset : The asset object to retrieve port and carrier information from

        Returns
        -------
        Tuple with the supply and return temperature.
        """

        assert len(asset.in_ports) <= 2 and len(asset.out_ports) == 1

        for port in asset.in_ports:
            if isinstance(port.carrier, esdl.HeatCommodity):
                in_carrier = asset.global_properties["carriers"][port.carrier.id]
        out_carrier = asset.global_properties["carriers"][asset.out_ports[0].carrier.id]

        if in_carrier["id"] == out_carrier["id"]:
            # these are the pipes, nodes, valves, pumps
            modifiers = {
                "temperature": in_carrier["temperature"],
                "carrier_id": in_carrier["id_number_mapping"],
            }
        else:
            # These are the sources, storages and consumers
            supply_temperature = (
                in_carrier["temperature"]
                if in_carrier["temperature"] > out_carrier["temperature"]
                else out_carrier["temperature"]
            )
            return_temperature = (
                in_carrier["temperature"]
                if in_carrier["temperature"] < out_carrier["temperature"]
                else out_carrier["temperature"]
            )
            temperature_supply_id = (
                in_carrier["id_number_mapping"]
                if in_carrier["temperature"] > out_carrier["temperature"]
                else out_carrier["id_number_mapping"]
            )
            temperature_return_id = (
                in_carrier["id_number_mapping"]
                if in_carrier["temperature"] < out_carrier["temperature"]
                else out_carrier["id_number_mapping"]
            )

            modifiers = {
                "T_supply": supply_temperature,
                "T_return": return_temperature,
                "T_supply_id": temperature_supply_id,
                "T_return_id": temperature_return_id,
            }
        return modifiers

    def _supply_return_temperature_modifiers(self, asset: Asset) -> MODIFIERS:
        """
        This function returns a dict containing all relevant temperatures associated with the asset
        needed for the optimization. These are the temperatures of the carrier at the inport and
        outport.

        Parameters
        ----------
        asset : Asset object to retrieve carrier temperatures from.

        Returns
        -------
        dict with all the temperatures.
        """

        if len(asset.in_ports) <= 2 and len(asset.out_ports) == 1:
            modifiers = self._get_supply_return_temperatures(asset)
            return modifiers
        elif len(asset.in_ports) >= 2 and len(asset.out_ports) == 2:
            prim_return_temperature = None
            sec_return_temperature = None
            for p in asset.in_ports:
                if isinstance(p.carrier, esdl.HeatCommodity):
                    carrier = asset.global_properties["carriers"][p.carrier.id]
                    if self.secondary_port_name_convention in p.name.lower():
                        sec_return_temperature_id = carrier["id_number_mapping"]
                        sec_return_temperature = carrier["temperature"]
                    else:
                        prim_supply_temperature = carrier["temperature"]
                        prim_supply_temperature_id = carrier["id_number_mapping"]
            for p in asset.out_ports:
                if isinstance(p.carrier, esdl.HeatCommodity):
                    carrier = asset.global_properties["carriers"][p.carrier.id]
                    if self.primary_port_name_convention in p.name.lower():
                        prim_return_temperature_id = carrier["id_number_mapping"]
                        prim_return_temperature = carrier["temperature"]
                    else:
                        sec_supply_temperature_id = carrier["id_number_mapping"]
                        sec_supply_temperature = carrier["temperature"]
            if not prim_return_temperature or not sec_return_temperature:
                raise RuntimeError(
                    f"{asset.name} ports are not specified correctly there should be dedicated "
                    f"primary and secondary ports ('prim' and 'sec') for the hydraulically "
                    f"decoupled networks"
                )
            assert sec_supply_temperature > sec_return_temperature
            assert sec_return_temperature > 0.0
            assert prim_supply_temperature > prim_return_temperature
            assert prim_return_temperature > 0.0
            temperatures = {
                "Primary": {
                    "T_supply": prim_supply_temperature,
                    "T_return": prim_return_temperature,
                    "T_supply_id": prim_supply_temperature_id,
                    "T_return_id": prim_return_temperature_id,
                },
                "Secondary": {
                    "T_supply": sec_supply_temperature,
                    "T_return": sec_return_temperature,
                    "T_supply_id": sec_supply_temperature_id,
                    "T_return_id": sec_return_temperature_id,
                },
            }
            return temperatures
        else:
            # unknown model type
            return {}

    @staticmethod
    def get_state(asset: Asset) -> float:
        """
        This function returns a float value, which represents the state (Enabled/disabled/optional)
        of an asset, so that it can be stored in the parameters.

        Parameters
        ----------
        asset : The asset object for retrieving the state

        Returns
        -------
        float value representing the asset's state
        """

        if asset.attributes["state"].name == "DISABLED":
            value = 0.0
        elif asset.attributes["state"].name == "OPTIONAL":
            value = 2.0
        else:
            value = 1.0
        return value

    def get_variable_opex_costs(self, asset: Asset) -> float:
        """
        Returns the variable opex costs coefficient of an asset in Euros per Wh.

        Parameters
        ----------
        asset : Asset object to get the cost information from

        Returns
        -------
        float for the variable operational cost coefficient.
        """

        cost_infos = dict()
        cost_infos["variableOperationalAndMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].variableOperationalAndMaintenanceCosts
        cost_infos["variableOperationalCosts"] = asset.attributes[
            "costInformation"
        ].variableOperationalCosts
        cost_infos["variableMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].variableMaintenanceCosts

        if all(cost_info is None for cost_info in cost_infos.values()):
            logger.warning(f"No variable OPEX cost information specified for asset {asset.name}")

        value = 0.0
        for cost_info in cost_infos.values():
            if cost_info is None:
                continue
            cost_value, unit, per_unit, per_time = self.get_cost_value_and_unit(cost_info)
            if unit != UnitEnum.EURO:
                logger.warning(f"Expected cost information {cost_info} to provide a cost in euros.")
                continue
            if per_time != TimeUnitEnum.NONE:
                logger.warning(
                    f"Specified OPEX for asset {asset.name} include a "
                    f"component per time, which we cannot handle."
                )
                continue
            if per_unit != UnitEnum.WATTHOUR and asset.asset_type not in [
                "GasDemand",
                "GasStorage",
                "Electrolyzer",
            ]:
                logger.warning(
                    f"Expected the specified OPEX for asset "
                    f"{asset.name} to be per Wh, but they are provided "
                    f"in {per_unit} instead."
                )
                continue
            if (
                asset.asset_type in ["GasDemand", "GasStorage", "Electrolyzer"]
                and per_unit != UnitEnum.GRAM
            ):
                logger.warning(
                    f"Expected the specified OPEX for asset "
                    f"{asset.name} to be per g/s, but they are provided "
                    f"in {per_unit}/{per_time} instead."
                )
                continue

            value += cost_value

        return value

    def get_fixed_opex_costs(self, asset: Asset) -> float:
        """
        Returns the fixed opex cost coefficient of an asset in Euros per W.

        Parameters
        ----------
        asset : Asset object to retrieve cost information from

        Returns
        -------
        fixed operational cost coefficient.
        """
        cost_infos = dict()
        cost_infos["fixedOperationalAndMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].fixedOperationalAndMaintenanceCosts
        cost_infos["fixedOperationalCosts"] = asset.attributes[
            "costInformation"
        ].fixedOperationalCosts
        cost_infos["fixedMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].fixedMaintenanceCosts

        if all(cost_info is None for cost_info in cost_infos.values()):
            logger.warning(f"No fixed OPEX cost information specified for asset {asset.name}")
            value = 0.0
        else:
            value = 0.0
            for cost_info in cost_infos.values():
                if cost_info is None:
                    continue
                cost_value, unit, per_unit, per_time = self.get_cost_value_and_unit(cost_info)
                if unit != UnitEnum.EURO:
                    RuntimeWarning(
                        f"Expected cost information {cost_info} to " f"provide a cost in euros."
                    )
                    continue
                if per_unit == UnitEnum.CUBIC_METRE and asset.asset_type != "GasStorage":
                    # index is 0 because buffers only have one in out port
                    supply_temp = asset.global_properties["carriers"][asset.in_ports[0].carrier.id][
                        "temperature"
                    ]
                    return_temp = asset.global_properties["carriers"][
                        asset.out_ports[0].carrier.id
                    ]["temperature"]
                    delta_temp = supply_temp - return_temp
                    m3_to_joule_factor = delta_temp * HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS
                    cost_value = cost_value / m3_to_joule_factor
                elif per_unit == UnitEnum.NONE:
                    if asset.asset_type == "HeatStorage":
                        size = asset.attributes["capacity"]
                        if size == 0.0:
                            # index is 0 because buffers only have one in out port
                            supply_temp = asset.global_properties["carriers"][
                                asset.in_ports[0].carrier.id
                            ]["temperature"]
                            return_temp = asset.global_properties["carriers"][
                                asset.out_ports[0].carrier.id
                            ]["temperature"]
                            delta_temp = supply_temp - return_temp
                            m3_to_joule_factor = (
                                delta_temp * HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS
                            )
                            size = asset.attributes["volume"] * m3_to_joule_factor
                            if size == 0.0:
                                RuntimeWarning(f"{asset.name} has not capacity or volume set")
                                return 0.0
                    elif asset.asset_type == "ATES":
                        size = asset.attributes["maxChargeRate"]
                        if size == 0.0:
                            size = asset.attributes["capacity"] / (
                                365 * 24 * 3600 / 2
                            )  # only half a year it can load
                            if size == 0.0:
                                RuntimeWarning(
                                    f"{asset.name} has not capacity or maximum charge rate set"
                                )
                                return 0.0
                    else:
                        try:
                            size = asset.attributes["power"]
                            if size == 0.0:
                                continue
                        except KeyError:
                            return 0.0
                    cost_value = cost_value / size
                elif per_unit != UnitEnum.WATT and asset.asset_type != "GasStorage":
                    RuntimeWarning(
                        f"Expected the specified OPEX for asset "
                        f"{asset.name} to be per W or m3, but they are provided "
                        f"in {per_unit} instead."
                    )
                    continue
                # still to decide if the cost is per kg or per m3
                elif per_unit != UnitEnum.GRAM and asset.asset_type == "GasStorage":
                    RuntimeWarning(
                        f"Expected the specified OPEX for asset "
                        f"{asset.name} to be per GRAM, but they are provided "
                        f"in {per_unit} instead."
                    )
                    continue

                value += cost_value
        return value

    @staticmethod
    def get_cost_value_and_unit(cost_info: esdl.SingleValue) -> Tuple[float, Any, Any, Any]:
        """
        This function returns the cost coefficient with unit information thereof.

        Parameters
        ----------
        cost_info : The single value object with the float and unit info.

        Returns
        -------
        The value with the unit decomposed.
        """

        cost_value = cost_info.value
        unit_info = cost_info.profileQuantityAndUnit
        unit = unit_info.unit
        per_time_uni = unit_info.perTimeUnit
        per_unit = unit_info.perUnit
        multiplier = unit_info.multiplier
        per_multiplier = unit_info.perMultiplier

        cost_value *= MULTI_ENUM_NAME_TO_FACTOR[multiplier]
        cost_value /= MULTI_ENUM_NAME_TO_FACTOR[per_multiplier]

        return cost_value, unit, per_unit, per_time_uni

    def get_installation_costs(self, asset: Asset) -> float:
        """
        This function return the installation cost coefficient in EUR for a single aggregation
        count.

        Parameters
        ----------
        asset : The asset object for retrieving the cost information from.

        Returns
        -------
        A float with the installation cost coefficient.
        """

        cost_info = asset.attributes["costInformation"].installationCosts
        if cost_info is None:
            logger.warning(f"No installation cost info provided for asset " f"{asset.name}.")
            return 0.0
        cost_value, unit, per_unit, per_time = self.get_cost_value_and_unit(cost_info)
        if unit != UnitEnum.EURO:
            logger.warning(f"Expect cost information {cost_info} to " f"provide a cost in euros")
            return 0.0
        if not per_time == TimeUnitEnum.NONE:
            logger.warning(
                f"Specified installation costs of asset {asset.name}"
                f" include a component per time, which we "
                f"cannot handle."
            )
            return 0.0
        if not per_unit == UnitEnum.NONE:
            logger.warning(
                f"Specified installation costs of asset {asset.name}"
                f" include a component per unit {per_unit}, which we "
                f"cannot handle."
            )
            return 0.0
        return cost_value

    def get_investment_costs(self, asset: Asset, per_unit: UnitEnum = UnitEnum.WATT) -> float:
        """
        Returns the investment cost coefficient of an asset in Euros per size unit (mostly W).

        Parameters
        ----------
        asset : The asset object to retrieve the cost information from.
        per_unit : The per unit needed in the optimization, as this may differ for some assets
        like the buffer where it scales with volume instead of power.

        Returns
        -------
        float for the investment cost coefficient.
        """

        cost_info = asset.attributes["costInformation"].investmentCosts
        if cost_info is None:
            RuntimeWarning(f"No investment costs provided for asset " f"{asset.name}.")
            return 0.0
        (
            cost_value,
            unit_provided,
            per_unit_provided,
            per_time_provided,
        ) = self.get_cost_value_and_unit(cost_info)
        if unit_provided != UnitEnum.EURO:
            logger.warning(f"Expect cost information {cost_info} to " f"provide a cost in euros")
            return 0.0
        if not per_time_provided == TimeUnitEnum.NONE:
            logger.warning(
                f"Specified investment costs for asset {asset.name}"
                f" include a component per time, which we "
                f"cannot handle."
            )
            return 0.0
        if per_unit == UnitEnum.WATT:
            if not per_unit_provided == UnitEnum.WATT:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per W, but they "
                    f"are provided in {per_unit_provided} "
                    f"instead."
                )
            return cost_value
        elif per_unit == UnitEnum.WATTHOUR:
            if not per_unit_provided == UnitEnum.WATTHOUR:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per Wh, but they "
                    f"are provided in {per_unit_provided} "
                    f"instead."
                )
                return 0.0
            return cost_value
        elif per_unit == UnitEnum.METRE:
            if not per_unit_provided == UnitEnum.METRE:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per meter, but they "
                    f"are provided in {per_unit_provided} "
                    f"instead."
                )
                return 0.0
            return cost_value
        elif per_unit == UnitEnum.JOULE:
            if per_unit_provided == UnitEnum.WATTHOUR:
                return cost_value / WATTHOUR_TO_JOULE
            elif per_unit_provided == UnitEnum.CUBIC_METRE:
                # index is 0 because buffers only have one in out port
                supply_temp = asset.global_properties["carriers"][asset.in_ports[0].carrier.id][
                    "temperature"
                ]
                return_temp = asset.global_properties["carriers"][asset.out_ports[0].carrier.id][
                    "temperature"
                ]
                delta_temp = supply_temp - return_temp
                m3_to_joule_factor = delta_temp * HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS
                return cost_value / m3_to_joule_factor
            else:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per Wh or m3, but "
                    f"they are provided in {per_unit_provided} "
                    f"instead."
                )
                return 0.0
        else:
            logger.warning(
                f"Cannot provide investment costs for asset " f"{asset.name} per {per_unit}"
            )
            return 0.0



from mesido.pycml import Variable

from numpy import nan

from ._non_storage_component import _NonStorageComponent


class HeatSource(_NonStorageComponent):
    """
    The source component is there to insert thermal power (Heat) into the network.

    The heat to discharge constraints are set in the HeatMixin. We enforce that the outgoing
    temperature of the source matches the absolute thermal power, Q * cp * rho * T_sup == Heat,
    similar as with the demands. This allows us to guarantee that the flow can always carry, as
    the heat losses further downstream in the network are over-estimated with T_ret where in
    reality this temperature drops. It also implicitly assumes that the temperature drops in the
    network are small and thus satisfy minimum temperature requirements.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(
                    HeatIn=dict(Heat=dict(min=0.0)),  # ensures it is not cooled down to below 0C
                    HeatOut=dict(Heat=dict(min=0.0)),
                ),
                modifiers,
            ),
        )

        self.component_type = "heat_source"

        self.price = nan  # TODO: delete not needed anymore
        self.co2_coeff = 1.0
        self.pump_efficiency = 0.5

        # Assumption: heat in/out and added is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(Variable, "Heat_source", min=0.0, nominal=self.Heat_nominal)
        self.add_variable(Variable, "Emission", min=0.0, nominal=self.Heat_nominal)
        self.add_variable(Variable, "dH", min=0.0)
        self.add_variable(
            Variable, "Pump_power", min=0.0, nominal=self.Q_nominal * self.nominal_pressure
        )

        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))
        self.add_equation(
            (self.Pump_power - (self.HeatOut.Hydraulic_power - self.HeatIn.Hydraulic_power))
            / (self.Q_nominal * self.nominal_pressure)
        )

        self.add_equation(
            (self.HeatOut.Heat - (self.HeatIn.Heat + self.Heat_source)) / self.Heat_nominal
        )
        # self.add_equation(
        #     (self.Emission - self.Heat_source * self.co2_coeff)
        #     / (self.Heat_nominal * self.co2_coeff)
        # )

        self.add_equation((self.Heat_flow - self.Heat_source) / self.Heat_nominal)



import logging
from abc import abstractmethod

import casadi as ca

from mesido.base_component_type_mixin import BaseComponentTypeMixin

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.timeseries import Timeseries

logger = logging.getLogger("mesido")


class FinancialMixin(BaseComponentTypeMixin, CollocatedIntegratedOptimizationProblem):
    """
    The FinancialMixin is used to instantiate variables for the different cost components of the
    assets in the energy network and to set constraints to compute them based upon the usage and
    size of the assets.

    The FinancialMixin is logically dependent on the AssetSizingMixin, as for example investment
    cost depends on asset size. We used abstractmethods to make these classes independent of each
    other (one does not need to inherit the other). The TechnoEconomicMixin in now responsible to
    make sure that the methods are there to allow the FinancialMixin to call for example asset size
    variables. The main advantage is that we could now have different variations for the asset
    sizing functionality all using the same FinancialMixin.
    """

    def __init__(self, *args, **kwargs):
        """
        In this __init__ we prepare the dicts for the variables added by the HeatMixin class
        """

        # Variable for fixed operational cost
        self._asset_fixed_operational_cost_map = {}
        self.__asset_fixed_operational_cost_var = {}
        self.__asset_fixed_operational_cost_nominals = {}
        self.__asset_fixed_operational_cost_bounds = {}

        # Variable for variable operational cost
        self._asset_variable_operational_cost_map = {}
        self.__asset_variable_operational_cost_var = {}
        self.__asset_variable_operational_cost_bounds = {}
        self.__asset_variable_operational_cost_nominals = {}

        # Variable for investment cost
        self._asset_investment_cost_map = {}
        self.__asset_investment_cost_var = {}
        self.__asset_investment_cost_nominals = {}
        self.__asset_investment_cost_bounds = {}

        # Variable for installation cost
        self._asset_installation_cost_map = {}
        self.__asset_installation_cost_var = {}
        self.__asset_installation_cost_bounds = {}
        self.__asset_installation_cost_nominals = {}

        # Variable for the cumulative investment and installation cost made per asset
        self.__cumulative_investments_made_in_eur_map = {}
        self.__cumulative_investments_made_in_eur_var = {}
        self.__cumulative_investments_made_in_eur_nominals = {}
        self.__cumulative_investments_made_in_eur_bounds = {}

        # Variable for when in time an asset is realized
        self.__asset_is_realized_map = {}
        self.__asset_is_realized_var = {}
        self.__asset_is_realized_bounds = {}

        # Variable for annualized capex cost
        self._annualized_capex_var_map = {}
        self.__annualized_capex_var = {}
        self.__annualized_capex_var_bounds = {}
        self.__annualized_capex_var_nominals = {}

        # Variable for realized revenue
        self._asset_revenue_map = {}
        self.__asset_revenue_var = {}
        self.__asset_revenue_nominals = {}
        self.__asset_revenue_bounds = {}

        super().__init__(*args, **kwargs)

    def pre(self):
        """
        In this pre method we fill the dicts initiated in the __init__. This means that we create
        the Casadi variables and determine the bounds, nominals and create maps for easier
        retrieving of the variables.
        """
        super().pre()

        options = self.energy_system_options()
        parameters = self.parameters(0)
        bounds = self.bounds()

        # Making the cost variables; fixed_operational_cost, variable_operational_cost,
        # installation_cost and investment_cost
        for asset_name in [
            asset_name
            for asset_name_list in self.energy_system_components.values()
            for asset_name in asset_name_list
        ]:
            if asset_name in [
                *self.energy_system_components.get("node", []),
                *self.energy_system_components.get("pump", []),
                *self.energy_system_components.get("check_valve", []),
                *self.energy_system_components.get("control_valve", []),
                *self.energy_system_components.get("electricity_node", []),
                *self.energy_system_components.get("gas_node", []),
            ]:
                continue
            elif asset_name in [*self.energy_system_components.get("ates", [])]:
                nominal_fixed_operational = self.variable_nominal(f"{asset_name}.Heat_ates")
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("low_temperature_ates", [])]:
                nominal_fixed_operational = self.variable_nominal(
                    f"{asset_name}.Heat_low_temperature_ates"
                )
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("heat_demand", [])]:
                nominal_fixed_operational = (
                    bounds[f"{asset_name}.Heat_demand"][1]
                    if not np.isinf(bounds[f"{asset_name}.Heat_demand"][1])
                    else bounds[f"{asset_name}.HeatIn.Heat"][1]
                )
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("heat_source", [])]:
                nominal_fixed_operational = self.variable_nominal(f"{asset_name}.Heat_source")
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("heat_pipe", [])]:
                nominal_fixed_operational = max(parameters[f"{asset_name}.length"], 1.0)
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("electricity_cable", [])]:
                nominal_fixed_operational = max(parameters[f"{asset_name}.length"], 1.0)
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("gas_pipe", [])]:
                nominal_fixed_operational = max(parameters[f"{asset_name}.length"], 1.0)
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("heat_buffer", [])]:
                nominal_fixed_operational = self.variable_nominal(f"{asset_name}.Stored_heat")
                nominal_variable_operational = self.variable_nominal(f"{asset_name}.Heat_buffer")
                nominal_investment = nominal_fixed_operational
            elif asset_name in [
                *self.energy_system_components.get("heat_exchanger", []),
                *self.energy_system_components.get("heat_pump", []),
            ]:
                nominal_fixed_operational = self.variable_nominal(f"{asset_name}.Secondary_heat")
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            # TODO: set the nominal values below
            elif asset_name in [*self.energy_system_components.get("gas_tank_storage", [])]:
                nominal_fixed_operational = bounds[f"{asset_name}.Stored_gas_mass"][1]
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("electricity_demand", [])]:
                nominal_fixed_operational = bounds[f"{asset_name}.Electricity_demand"][1]
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("electrolyzer", [])]:
                nominal_fixed_operational = bounds[f"{asset_name}.Power_consumed"][1]
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("electricity_source", [])]:
                max_power = (
                    bounds[f"{asset_name}.ElectricityOut.Power"][1]
                    if (isinstance(bounds[f"{asset_name}.ElectricityOut.Power"][1], float))
                    else max(bounds[f"{asset_name}.ElectricityOut.Power"][1].values)
                )
                nominal_fixed_operational = max_power
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("electricity_storage", [])]:
                nominal_fixed_operational = bounds[f"{asset_name}.Stored_electricity"][1]
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            elif asset_name in [*self.energy_system_components.get("gas_demand", [])]:
                nominal_fixed_operational = bounds[f"{asset_name}.Gas_demand_mass_flow"][1]
                nominal_variable_operational = nominal_fixed_operational
                nominal_investment = nominal_fixed_operational
            else:
                logger.warning(
                    f"Asset {asset_name} has type for which "
                    f"we cannot determine bounds and nominals on the costs, "
                    f"skipping it."
                )
                nominal_fixed_operational = 1.0
                nominal_variable_operational = 1.0
                nominal_investment = 1.0

            # fixed operational cost
            asset_fixed_operational_cost_var = f"{asset_name}__fixed_operational_cost"
            self._asset_fixed_operational_cost_map[asset_name] = asset_fixed_operational_cost_var
            self.__asset_fixed_operational_cost_var[asset_fixed_operational_cost_var] = ca.MX.sym(
                asset_fixed_operational_cost_var
            )
            self.__asset_fixed_operational_cost_bounds[asset_fixed_operational_cost_var] = (
                0.0,
                np.inf,
            )
            self.__asset_fixed_operational_cost_nominals[asset_fixed_operational_cost_var] = (
                max(
                    parameters[f"{asset_name}.fixed_operational_cost_coefficient"]
                    * nominal_fixed_operational,
                    1.0e2,
                )
                if nominal_fixed_operational is not None
                else 1.0e2
            )

            # variable operational cost
            variable_operational_cost_var = f"{asset_name}__variable_operational_cost"
            self._asset_variable_operational_cost_map[asset_name] = variable_operational_cost_var
            self.__asset_variable_operational_cost_var[variable_operational_cost_var] = ca.MX.sym(
                variable_operational_cost_var
            )
            self.__asset_variable_operational_cost_bounds[variable_operational_cost_var] = (
                0.0,
                np.inf,
            )
            self.__asset_variable_operational_cost_nominals[variable_operational_cost_var] = (
                max(
                    parameters[f"{asset_name}.variable_operational_cost_coefficient"]
                    * nominal_variable_operational
                    * 24.0,
                    1.0e2,
                )
                if nominal_variable_operational is not None
                else 1.0e2
            )

            # installation cost
            asset_installation_cost_var = f"{asset_name}__installation_cost"
            self._asset_installation_cost_map[asset_name] = asset_installation_cost_var
            self.__asset_installation_cost_var[asset_installation_cost_var] = ca.MX.sym(
                asset_installation_cost_var
            )
            try:
                aggr_count_max = parameters[f"{asset_name}.nr_of_doublets"]
            except KeyError:
                aggr_count_max = 1.0
            if parameters[f"{asset_name}.state"] == 0:
                aggr_count_max = 0.0
            self.__asset_installation_cost_bounds[asset_installation_cost_var] = (
                0.0,
                parameters[f"{asset_name}.installation_cost"] * aggr_count_max,
            )
            self.__asset_installation_cost_nominals[asset_installation_cost_var] = (
                parameters[f"{asset_name}.installation_cost"]
                if parameters[f"{asset_name}.installation_cost"]
                else 1.0e2
            )

            # investment cost
            asset_investment_cost_var = f"{asset_name}__investment_cost"
            self._asset_investment_cost_map[asset_name] = asset_investment_cost_var
            self.__asset_investment_cost_var[asset_investment_cost_var] = ca.MX.sym(
                asset_investment_cost_var
            )

            if asset_name in self.energy_system_components.get("heat_pipe", []):
                if asset_name in self.get_pipe_class_map().keys():
                    pipe_classes = self.get_pipe_class_map()[asset_name]
                    max_cost = (
                        2.0
                        * parameters[f"{asset_name}.length"]
                        * max([c.investment_costs for c in pipe_classes.keys()])
                    )
                else:
                    max_cost = (
                        2.0
                        * parameters[f"{asset_name}.length"]
                        * parameters[f"{asset_name}.investment_cost_coefficient"]
                    )
            elif asset_name in self.energy_system_components.get("gas_pipe", []):
                if asset_name in self.get_gas_pipe_class_map().keys():
                    pipe_classes = self.get_gas_pipe_class_map()[asset_name]
                    max_cost = (
                        2.0
                        * parameters[f"{asset_name}.length"]
                        * max([c.investment_costs for c in pipe_classes.keys()])
                    )
                else:
                    max_cost = (
                        2.0
                        * parameters[f"{asset_name}.length"]
                        * parameters[f"{asset_name}.investment_cost_coefficient"]
                    )
            elif asset_name in self.energy_system_components.get("electricity_cable", []):
                if asset_name in self.get_electricity_cable_class_map().keys():
                    cable_classes = self.get_electricity_cable_class_map()[asset_name]
                    max_cost = (
                        2.0
                        * parameters[f"{asset_name}.length"]
                        * max([c.investment_costs for c in cable_classes.keys()])
                    )
                else:
                    max_cost = (
                        2.0
                        * parameters[f"{asset_name}.length"]
                        * parameters[f"{asset_name}.investment_cost_coefficient"]
                    )
            else:
                max_cost = (
                    max(bounds[f"{asset_name}__max_size"][1].values)
                    * parameters[f"{asset_name}.investment_cost_coefficient"]
                    if isinstance(bounds[f"{asset_name}__max_size"][1], Timeseries)
                    else bounds[f"{asset_name}__max_size"][1]
                    * parameters[f"{asset_name}.investment_cost_coefficient"]
                )
            self.__asset_investment_cost_bounds[asset_investment_cost_var] = (0.0, max_cost)
            self.__asset_investment_cost_nominals[asset_investment_cost_var] = (
                max(
                    parameters[f"{asset_name}.investment_cost_coefficient"] * nominal_investment,
                    1.0e2,
                )
                if nominal_investment is not None
                else 1.0e2
            )

            # Realized revenue
            if (asset_name) in [
                *self.energy_system_components.get("electricity_demand", []),
                *self.energy_system_components.get("gas_demand", []),
            ]:

                carrier_name = None
                for _id, attr in self.get_electricity_carriers().items():
                    if attr["id_number_mapping"] == parameters[f"{asset_name}.id_mapping_carrier"]:
                        carrier_name = attr["name"]
                for _id, attr in self.get_gas_carriers().items():
                    if attr["id_number_mapping"] == parameters[f"{asset_name}.id_mapping_carrier"]:
                        carrier_name = attr["name"]
                if carrier_name is not None:
                    asset_revenue_var = f"{asset_name}__revenue"
                    self._asset_revenue_map[asset_name] = asset_revenue_var
                    self.__asset_revenue_var[asset_revenue_var] = ca.MX.sym(asset_revenue_var)
                    self.__asset_revenue_bounds[asset_revenue_var] = (
                        0.0,
                        np.inf,
                    )
                    self.__asset_revenue_nominals[asset_revenue_var] = (
                        max(
                            np.mean(self.get_timeseries(f"{carrier_name}.price_profile").values)
                            * nominal_fixed_operational,
                            1.0e2,
                        )
                        if nominal_fixed_operational is not None
                        else 1.0e2
                    )

        for asset in [
            *self.energy_system_components.get("heat_source", []),
            *self.energy_system_components.get("heat_demand", []),
            *self.energy_system_components.get("ates", []),
            *self.energy_system_components.get("low_temperature_ates", []),
            *self.energy_system_components.get("heat_buffer", []),
            *self.energy_system_components.get("heat_pipe", []),
            *self.energy_system_components.get("heat_exchanger", []),
            *self.energy_system_components.get("heat_pump", []),
        ]:
            annualized_capex_var_name = f"{asset}__annualized_capex"
            self._annualized_capex_var_map[asset] = annualized_capex_var_name
            self.__annualized_capex_var[annualized_capex_var_name] = ca.MX.sym(
                annualized_capex_var_name
            )
            self.__annualized_capex_var_bounds[annualized_capex_var_name] = (
                0.0,
                np.inf,
            )  # (lb, ub)
            installation_cost_symbol_name = self._asset_installation_cost_map[asset]
            investment_cost_symbol_name = self._asset_investment_cost_map[asset]
            self.__annualized_capex_var_nominals[annualized_capex_var_name] = self.variable_nominal(
                installation_cost_symbol_name
            ) + self.variable_nominal(investment_cost_symbol_name)

        if options["include_asset_is_realized"]:
            for asset in [
                *self.energy_system_components.get("heat_source", []),
                *self.energy_system_components.get("heat_demand", []),
                *self.energy_system_components.get("ates", []),
                *self.energy_system_components.get("low_temperature_ates", []),
                *self.energy_system_components.get("heat_buffer", []),
                *self.energy_system_components.get("heat_exchanger", []),
                *self.energy_system_components.get("heat_pump", []),
            ]:
                var_name = f"{asset}__cumulative_investments_made_in_eur"
                self.__cumulative_investments_made_in_eur_map[asset] = var_name
                self.__cumulative_investments_made_in_eur_var[var_name] = ca.MX.sym(var_name)
                self.__cumulative_investments_made_in_eur_nominals[var_name] = (
                    self.variable_nominal(f"{asset}__investment_cost")
                    + self.variable_nominal(f"{asset}__installation_cost")
                )
                self.__cumulative_investments_made_in_eur_bounds[var_name] = (0.0, np.inf)

                # This is an integer variable between [0, max_aggregation_count] that allows the
                # increments of the asset to become used by the optimizer. Meaning that when this
                # variable is zero not milp can be consumed or produced by this asset. When the
                # integer is >=1 the asset can consume and/or produce according to its increments.
                var_name = f"{asset}__asset_is_realized"
                self.__asset_is_realized_map[asset] = var_name
                self.__asset_is_realized_var[var_name] = ca.MX.sym(var_name)
                try:
                    aggr_count_max = parameters[f"{asset}.nr_of_doublets"]
                except KeyError:
                    aggr_count_max = 1.0
                if parameters[f"{asset}.state"] == 0:
                    aggr_count_max = 0.0
                self.__asset_is_realized_bounds[var_name] = (0.0, aggr_count_max)

    @abstractmethod
    def energy_system_options(self):
        r"""
        Returns a dictionary of milp network specific options.

        +--------------------------------------+-----------+-----------------------------+
        | ``discounted_annualized_cost ``       | ``bool``  | ``False``                   |
        +--------------------------------------+-----------+-----------------------------+

        The ``discounted_annualized_cost`` option computes the annualized discounted costs for
        each asset, and defines the sum of these costs as the total cost of ownership for the
        cost minimization goal.
        """

        options = {}

        options["discounted_annualized_cost"] = False

        return options

    @abstractmethod
    def get_max_size_var(self, asset_name, ensemble_member):
        """
        This function should return the max size variable of an asset.

        Returns
        -------
        Casadi symbol for the maximum size variable
        """
        raise NotImplementedError

    @abstractmethod
    def get_aggregation_count_var(self, asset_name, ensemble_member):
        """
        This function should return the aggregation count integer variable of an asset.

        Returns
        -------
        Casadi symbol for the aggregation count variable
        """
        raise NotImplementedError

    @abstractmethod
    def get_aggregation_count_max(self, asset_name):
        """
        This function should return the aggregation count upper bound.

        Returns
        -------
        scalar or int with maximum amount of the aggregation count.
        """
        raise NotImplementedError

    @abstractmethod
    def get_pipe_investment_cost_coefficient(self, asset_name, ensemble_member):
        """
        This function should return the pipe investment cost coefficient variable.

        Returns
        -------
        Casadi symbol for the investment cost coefficient of the pipe.
        """
        raise NotImplementedError

    @abstractmethod
    def get_gas_pipe_investment_cost_coefficient(self, asset_name, ensemble_member):
        """
        This function should return the gas_pipe investment cost coefficient variable.

        Returns
        -------
        Casadi symbol for the investment cost coefficient of the pipe.
        """
        raise NotImplementedError

    @abstractmethod
    def get_electricity_cable_investment_cost_coefficient(self, asset_name, ensemble_member):
        """
        This function should return the electricity cable investment cost coefficient variable.

        Returns
        -------
        Casadi symbol for the investment cost coefficient of the cable.
        """
        raise NotImplementedError

    @abstractmethod
    def get_pipe_class_map(self):
        """
        This function should return the mapping between the pipe and all the possible pipe classes
        available for that pipe.

        Returns
        -------

        """
        return NotImplementedError

    @abstractmethod
    def get_electricity_carriers(self, type=None):
        """
        This function should return the mapping between the pipe and all the possible pipe classes
        available for that pipe.

        Returns
        -------

        """
        return NotImplementedError

    @abstractmethod
    def get_gas_pipe_class_map(self):
        """
        This function should return the mapping between the gas pipe and all the possible pipe
        classes available for that pipe.

        Returns
        -------

        """
        return NotImplementedError

    @abstractmethod
    def get_gas_carriers(self, type=None):
        """
        This function should return all the gas carriers

        Returns
        -------

        """
        return NotImplementedError

    @abstractmethod
    def get_heat_carriers(self, type=None):
        """
        This function should return all the heat carriers

        Returns
        -------

        """
        return NotImplementedError

    @abstractmethod
    def get_electricity_cable_class_map(self):
        """
        This function should return the mapping between the cable and all the possible cable classes
        available for that cable.

        Returns
        -------

        """
        return NotImplementedError

    @property
    def extra_variables(self):
        """
        In this function we add all the variables defined in the HeatMixin to the optimization
        problem. Note that these are only the normal variables not path variables.
        """
        variables = super().extra_variables.copy()
        variables.extend(self.__asset_fixed_operational_cost_var.values())
        variables.extend(self.__asset_investment_cost_var.values())
        variables.extend(self.__asset_installation_cost_var.values())
        variables.extend(self.__asset_variable_operational_cost_var.values())
        variables.extend(self.__annualized_capex_var.values())
        variables.extend(self.__asset_revenue_var.values())
        return variables

    @property
    def path_variables(self):
        """
        In this function we add all the path variables defined in the HeatMixin to the
        optimization problem. Note that path_variables are variables that are created for each
        time-step.
        """
        variables = super().path_variables.copy()
        variables.extend(self.__cumulative_investments_made_in_eur_var.values())
        variables.extend(self.__asset_is_realized_var.values())
        return variables

    def variable_is_discrete(self, variable):
        """
        All variables that only can take integer values should be added to this function.
        """
        if variable in self.__asset_is_realized_var:
            return True
        else:
            return super().variable_is_discrete(variable)

    def variable_nominal(self, variable):
        """
        In this function we add all the nominals for the variables defined/added in the HeatMixin.
        """
        if variable in self.__asset_fixed_operational_cost_nominals:
            return self.__asset_fixed_operational_cost_nominals[variable]
        elif variable in self.__asset_investment_cost_nominals:
            return self.__asset_investment_cost_nominals[variable]
        elif variable in self.__asset_variable_operational_cost_nominals:
            return self.__asset_variable_operational_cost_nominals[variable]
        elif variable in self.__asset_installation_cost_nominals:
            return self.__asset_installation_cost_nominals[variable]
        elif variable in self.__cumulative_investments_made_in_eur_nominals:
            return self.__cumulative_investments_made_in_eur_nominals[variable]
        elif variable in self.__annualized_capex_var_nominals:
            return self.__annualized_capex_var_nominals[variable]
        elif variable in self.__asset_revenue_nominals:
            return self.__asset_revenue_nominals[variable]
        else:
            return super().variable_nominal(variable)

    def bounds(self):
        """
        In this function we add the bounds to the problem for all the variables defined/added in
        the HeatMixin.
        """
        bounds = super().bounds()
        bounds.update(self.__asset_fixed_operational_cost_bounds)
        bounds.update(self.__asset_investment_cost_bounds)
        bounds.update(self.__asset_installation_cost_bounds)
        bounds.update(self.__asset_variable_operational_cost_bounds)
        bounds.update(self.__asset_is_realized_bounds)
        bounds.update(self.__cumulative_investments_made_in_eur_bounds)
        bounds.update(self.__annualized_capex_var_bounds)
        bounds.update(self.__asset_revenue_bounds)
        return bounds

    @staticmethod
    def __get_abs_max_bounds(*bounds):
        """
        This function returns the absolute maximum of the bounds given. Note that bounds can also be
        a timeseries.
        """
        max_ = 0.0

        for b in bounds:
            if isinstance(b, np.ndarray):
                max_ = max(max_, max(abs(b)))
            elif isinstance(b, Timeseries):
                max_ = max(max_, max(abs(b.values)))
            else:
                max_ = max(max_, abs(b))

        return max_

    def __state_vector_scaled(self, variable, ensemble_member):
        """
        This functions returns the casadi symbols scaled with their nominal for the entire time
        horizon.
        """
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def __investment_cost_constraints(self, ensemble_member):
        """
        This function adds constraints to set the investment cost variable. The investment cost
        scales with the maximum size of the asset. This leaves two cases for the constraint. 1) The
        asset size is fixed (state==1): in this case the investment cost is set based on the upper
        bound of the size. 2) The asset size is optimized (state==2): in this case the investment
        cost is set based upon the __max_size variable.

        Specifically for demands we have a case where we set the investment cost based on the
        maximum demand as often the size of the demand is not seperately specified.

        For pipes the investment cost is set based on the pipe class and the length.
        """
        constraints = []

        parameters = self.parameters(ensemble_member)

        for asset_name in [
            asset_name
            for asset_name_list in self.energy_system_components.values()
            for asset_name in asset_name_list
        ]:
            if asset_name in [
                *self.energy_system_components.get("node", []),
                *self.energy_system_components.get("pump", []),
                *self.energy_system_components.get("check_valve", []),
                *self.energy_system_components.get("electricity_node", []),
                *self.energy_system_components.get("gas_node", []),
                *self.energy_system_components.get("gas_tank_storage", []),
            ]:
                # TODO: add support for joints?
                continue

            investment_cost_var = self._asset_investment_cost_map[asset_name]
            investment_costs = self.extra_variable(investment_cost_var, ensemble_member)
            investment_cost_coefficient = parameters[f"{asset_name}.investment_cost_coefficient"]
            nominal = self.variable_nominal(investment_cost_var)

            if asset_name in [*self.energy_system_components.get("heat_pipe", [])]:
                # We do the pipe seperately as their coefficients are specified per meter.
                investment_cost_coefficient = self.get_pipe_investment_cost_coefficient(
                    asset_name, ensemble_member
                )
                asset_size = parameters[f"{asset_name}.length"]
            elif asset_name in [*self.energy_system_components.get("gas_pipe", [])]:
                # We do the pipe seperately as their coefficients are specified per meter.
                investment_cost_coefficient = self.get_gas_pipe_investment_cost_coefficient(
                    asset_name, ensemble_member
                )
                asset_size = parameters[f"{asset_name}.length"]
            elif asset_name in [*self.energy_system_components.get("electricity_cable", [])]:
                # We do the pipe seperately as their coefficients are specified per meter.
                investment_cost_coefficient = (
                    self.get_electricity_cable_investment_cost_coefficient(
                        asset_name, ensemble_member
                    )
                )
                asset_size = parameters[f"{asset_name}.length"]
            else:
                asset_size = self.get_max_size_var(asset_name, ensemble_member)

            constraints.append(
                (
                    (investment_costs - asset_size * investment_cost_coefficient) / nominal,
                    0.0,
                    0.0,
                )
            )

        return constraints

    def __fixed_operational_cost_constraints(self, ensemble_member):
        """
        This function adds the constraints to set the fixed operational cost. The fixed operational
        cost are the cost made independently of the operation of the asset. We assume that these
        cost scale with the maximum size of the asset.
        """
        constraints = []

        parameters = self.parameters(ensemble_member)

        for asset_name in [
            asset_name
            for asset_name_list in self.energy_system_components.values()
            for asset_name in asset_name_list
        ]:
            if asset_name in [
                *self.energy_system_components.get("node", []),
                *self.energy_system_components.get("heat_pipe", []),
                *self.energy_system_components.get("electricity_cable", []),
                *self.energy_system_components.get("electricity_node", []),
                *self.energy_system_components.get("gas_pipe", []),
                *self.energy_system_components.get("gas_node", []),
                *self.energy_system_components.get("pump", []),
                *self.energy_system_components.get("check_valve", []),
            ]:
                # currently no support for joints
                continue
            fixed_operational_cost_var = self._asset_fixed_operational_cost_map[asset_name]
            fixed_operational_cost = self.extra_variable(
                fixed_operational_cost_var, ensemble_member
            )
            asset_size = self.get_max_size_var(asset_name, ensemble_member)
            fixed_operational_cost_coefficient = parameters[
                f"{asset_name}.fixed_operational_cost_coefficient"
            ]

            nominal = self.variable_nominal(fixed_operational_cost_var)

            constraints.append(
                (
                    (fixed_operational_cost - asset_size * fixed_operational_cost_coefficient)
                    / nominal,
                    0.0,
                    0.0,
                )
            )

        return constraints

    def __variable_operational_cost_constraints(self, ensemble_member):
        """
        This function adds the constraints for setting the variable operational cost. These are the
        cost that depend on the operation of the asset. At this moment we only support the variable
        operational cost for sources where they scale with the thermal energy production.
        """
        constraints = []

        parameters = self.parameters(ensemble_member)

        for asset in [
            *self.energy_system_components.get("ates", []),
            *self.energy_system_components.get("low_temperature_ates", []),
            *self.energy_system_components.get("heat_buffer", []),
            *self.energy_system_components.get("pump", []),
            *self.energy_system_components.get("heat_exchanger", []),
        ]:
            variable_operational_cost_var = self._asset_variable_operational_cost_map[asset]
            variable_operational_cost = self.extra_variable(
                variable_operational_cost_var, ensemble_member
            )
            nominal = self.variable_nominal(variable_operational_cost_var)

            pump_power = self.__state_vector_scaled(f"{asset}.Pump_power", ensemble_member)
            eff = parameters[f"{asset}.pump_efficiency"]

            # We assume that only one electricity carrier is specified, to compute the cost with.
            # Otherwise we need to link the electricity carrier somehow to the source and pump asset
            # which is lots of extra effort for the user.
            assert len(self.get_electricity_carriers().keys()) <= 1

            if len(self.get_electricity_carriers().keys()) == 1:
                price_profile = self.get_timeseries(
                    f"{list(self.get_electricity_carriers().values())[0]['name']}.price_profile"
                )
            else:
                price_profile = Timeseries(self.times(), np.zeros(len(self.times())))

            timesteps = np.diff(self.times()) / 3600.0

            sum = 0.0
            for i in range(1, len(self.times())):
                sum += price_profile.values[i] * pump_power[i] * timesteps[i - 1] / eff

            constraints.append(((variable_operational_cost - sum) / nominal, 0.0, 0.0))

        for s in self.energy_system_components.get("heat_source", []):
            heat_source = self.__state_vector_scaled(f"{s}.Heat_source", ensemble_member)
            variable_operational_cost_var = self._asset_variable_operational_cost_map[s]
            variable_operational_cost = self.extra_variable(
                variable_operational_cost_var, ensemble_member
            )
            nominal = self.variable_nominal(variable_operational_cost_var)
            variable_operational_cost_coefficient = parameters[
                f"{s}.variable_operational_cost_coefficient"
            ]
            timesteps = np.diff(self.times()) / 3600.0

            pump_power = self.__state_vector_scaled(f"{s}.Pump_power", ensemble_member)
            eff = parameters[f"{s}.pump_efficiency"]

            # We assume that only one electricity carrier is specified, to compute the cost with.
            # Otherwise we need to link the electricity carrier somehow to the source and pump asset
            # which is lots of extra effort for the user.
            assert len(self.get_electricity_carriers().keys()) <= 1

            if len(self.get_electricity_carriers().keys()) == 1:
                price_profile = self.get_timeseries(
                    f"{list(self.get_electricity_carriers().values())[0]['name']}.price_profile"
                )
            else:
                price_profile = Timeseries(self.times(), np.zeros(len(self.times())))

            sum = 0.0
            for i in range(1, len(self.times())):
                sum += variable_operational_cost_coefficient * heat_source[i] * timesteps[i - 1]
                sum += price_profile.values[i] * pump_power[i] * timesteps[i - 1] / eff

            constraints.append(((variable_operational_cost - sum) / nominal, 0.0, 0.0))

        for hp in [
            *self.energy_system_components.get("heat_pump", []),
        ]:
            elec_consumption = self.__state_vector_scaled(f"{hp}.Power_elec", ensemble_member)
            variable_operational_cost_var = self._asset_variable_operational_cost_map[hp]
            variable_operational_cost = self.extra_variable(
                variable_operational_cost_var, ensemble_member
            )
            nominal = self.variable_nominal(variable_operational_cost_var)
            variable_operational_cost_coefficient = parameters[
                f"{hp}.variable_operational_cost_coefficient"
            ]
            timesteps = np.diff(self.times()) / 3600
            pump_power = self.__state_vector_scaled(f"{hp}.Pump_power", ensemble_member)
            eff = parameters[f"{hp}.pump_efficiency"]

            # We assume that only one electricity carrier is specified, to compute the cost with.
            # Otherwise we need to link the electricity carrier somehow to the source and pump asset
            # which is lots of extra effort for the user.
            assert len(self.get_electricity_carriers().keys()) <= 1

            if len(self.get_electricity_carriers().keys()) == 1:
                price_profile = self.get_timeseries(
                    f"{list(self.get_electricity_carriers().values())[0]['name']}.price_profile"
                )
            else:
                price_profile = Timeseries(self.times(), np.zeros(len(self.times())))

            sum = 0.0
            for i in range(1, len(self.times())):
                sum += (
                    variable_operational_cost_coefficient * elec_consumption[i] * timesteps[i - 1]
                )
                sum += price_profile.values[i] * pump_power[i] * timesteps[i - 1] / eff
                if hp not in self.energy_system_components.get("heat_pump_elec", []):
                    # assuming that if heatpump has electricity port, the cost for the electricity
                    # are already made by the electricity producer and transport
                    sum += price_profile.values[i] * elec_consumption[i] * timesteps[i - 1]

            constraints.append(((variable_operational_cost - sum) / nominal, 0.0, 0.0))

        for _ in self.energy_system_components.get("heat_buffer", []):
            pass

        for demand in self.energy_system_components.get("gas_demand", []):
            gas_mass_flow = self.__state_vector_scaled(
                f"{demand}.Gas_demand_mass_flow", ensemble_member  # g/s
            )

            variable_operational_cost_var = self._asset_variable_operational_cost_map[demand]
            variable_operational_cost = self.extra_variable(
                variable_operational_cost_var, ensemble_member
            )
            nominal = self.variable_nominal(variable_operational_cost_var)
            variable_operational_cost_coefficient = parameters[
                f"{demand}.variable_operational_cost_coefficient"
            ]

            sum = 0.0
            timesteps = np.diff(self.times()) / 3600.0
            for i in range(1, len(self.times())):
                sum += variable_operational_cost_coefficient * gas_mass_flow[i] * timesteps[i - 1]

            constraints.append(((variable_operational_cost - sum) / nominal, 0.0, 0.0))

        # for a in self.heat_network_components.get("ates", []):
        # TODO: needs to be replaced with the positive or abs value of this, see varOPEX,
        #  then ates varopex also needs to be added to the mnimize_tco_goal
        # heat_ates = self.__state_vector_scaled(f"{a}.Heat_ates", ensemble_member)
        # variable_operational_cost_var = self._asset_variable_operational_cost_map[a]
        # variable_operational_cost = self.extra_variable(
        #     variable_operational_cost_var, ensemble_member
        # )
        # nominal = self.variable_nominal(variable_operational_cost_var)
        # variable_operational_cost_coefficient = parameters[
        #     f"{a}.variable_operational_cost_coefficient"
        # ]
        # timesteps = np.diff(self.times()) / 3600.0
        #
        # sum = 0.0
        #
        # for i in range(1, len(self.times())):
        #     varOPEX_dt = (variable_operational_cost_coefficient * heat_ates[i]
        #     * timesteps[i - 1])
        #     constraints.append(((varOPEX-varOPEX_dt)/nominal,0.0, np,inf))
        #     #varOPEX would be a variable>0 for everyt timestep
        #     sum += varOPEX
        # constraints.append(((variable_operational_cost - sum) / (nominal), 0.0, 0.0))

        for electrolyzer in self.energy_system_components.get("electrolyzer", []):
            power_consumer = self.__state_vector_scaled(
                f"{electrolyzer}.Gas_mass_flow_out", ensemble_member
            )

            variable_operational_cost_var = self._asset_variable_operational_cost_map[electrolyzer]
            variable_operational_cost = self.extra_variable(
                variable_operational_cost_var, ensemble_member
            )
            nominal = self.variable_nominal(variable_operational_cost_var)
            variable_operational_cost_coefficient = parameters[
                f"{electrolyzer}.variable_operational_cost_coefficient"
            ]

            sum = 0.0
            timesteps = np.diff(self.times()) / 3600.0
            for i in range(1, len(self.times())):
                sum += (
                    variable_operational_cost_coefficient
                    * power_consumer[i]
                    * timesteps[i - 1]  # gas_mass_flow unit is g/s
                )

            constraints.append(((variable_operational_cost - sum) / nominal, 0.0, 0.0))

        # for a in self.heat_network_components.get("ates", []):
        # TODO: needs to be replaced with the positive or abs value of this, see varOPEX,
        #  then ates varopex also needs to be added to the mnimize_tco_goal
        # heat_ates = self.__state_vector_scaled(f"{a}.Heat_ates", ensemble_member)
        # variable_operational_cost_var = self._asset_variable_operational_cost_map[a]
        # variable_operational_cost = self.extra_variable(
        #     variable_operational_cost_var, ensemble_member
        # )
        # nominal = self.variable_nominal(variable_operational_cost_var)
        # variable_operational_cost_coefficient = parameters[
        #     f"{a}.variable_operational_cost_coefficient"
        # ]
        # timesteps = np.diff(self.times()) / 3600.0
        #
        # sum = 0.0
        #
        # for i in range(1, len(self.times())):
        #     varOPEX_dt = (variable_operational_cost_coefficient * heat_ates[i]
        #     * timesteps[i - 1])
        #     constraints.append(((varOPEX-varOPEX_dt)/nominal,0.0, np,inf))
        #     #varOPEX would be a variable>0 for everyt timestep
        #     sum += varOPEX
        # constraints.append(((variable_operational_cost - sum) / (nominal), 0.0, 0.0))

        return constraints

    def __installation_cost_constraints(self, ensemble_member):
        """
        This function adds the constraints for setting the installation cost variable. The
        installation cost is the cost element that comes with the placing of the asset
        independently of the size of the asset. Therefore, the installation cost is set with the
        _aggregation_count variable.
        """
        constraints = []

        parameters = self.parameters(ensemble_member)

        for asset_name in [
            asset_name
            for asset_name_list in self.energy_system_components.values()
            for asset_name in asset_name_list
        ]:
            if asset_name in [
                *self.energy_system_components.get("node", []),
                *self.energy_system_components.get("pump", []),
                *self.energy_system_components.get("check_valve", []),
                *self.energy_system_components.get("electricity_cable", []),
                *self.energy_system_components.get("electricity_node", []),
                *self.energy_system_components.get("gas_pipe", []),
                *self.energy_system_components.get("gas_node", []),
            ]:
                # no support for joints right now
                continue
            installation_cost_sym = self.extra_variable(
                self._asset_installation_cost_map[asset_name]
            )
            nominal = self.variable_nominal(self._asset_installation_cost_map[asset_name])
            installation_cost = parameters[f"{asset_name}.installation_cost"]
            aggregation_count_sym = self.get_aggregation_count_var(asset_name, ensemble_member)
            constraints.append(
                (
                    (installation_cost_sym - aggregation_count_sym * installation_cost) / nominal,
                    0.0,
                    0.0,
                )
            )

        return constraints

    def __cumulative_investments_made_in_eur_path_constraints(self, ensemble_member):
        r"""
        These constraints are linking the cummulitive investments made to the possiblity of
        utilizing an asset. The investments made are sufficient for that asset to be realized it
        becomes available.

        Meaning that an asset requires 1million euro investment to be realized
        and the investments made at timestep i are sufficient the asset also is realized (becomes
        available) in that same timestep.
        """
        constraints = []
        options = self.energy_system_options()
        if options["include_asset_is_realized"]:
            for asset in [
                # *self.energy_system_components.get("heat_demand", []),
                *self.energy_system_components.get("heat_source", []),
                *self.energy_system_components.get("ates", []),
                *self.energy_system_components.get("low_temperature_ates", []),
                *self.energy_system_components.get("heat_buffer", []),
                *self.energy_system_components.get("heat_exchanger", []),
                *self.energy_system_components.get("heat_pump", []),
            ]:
                var_name = self.__cumulative_investments_made_in_eur_map[asset]
                cumulative_investments_made = self.state(var_name)
                nominal = self.variable_nominal(var_name)
                var_name = self.__asset_is_realized_map[asset]
                asset_is_realized = self.state(var_name)
                installation_cost_sym = self.variable(self._asset_installation_cost_map[asset])
                investment_cost_sym = self.variable(self._asset_investment_cost_map[asset])
                # TODO: add insulation class cost to the investments made.
                # if asset in self.heat_network_components.get("demand", []):
                #     for insulation_class in self.__get_insulation_classes(asset):
                #         insulation_class_active
                #         insulation_class_cost
                #         investment_cost_sym += insulation_class_active * insulation_class_cost
                big_m = (
                    1.5
                    * max(
                        self.bounds()[f"{asset}__investment_cost"][1]
                        + self.bounds()[f"{asset}__installation_cost"][1],
                        1.0,
                    )
                    / max(self.get_aggregation_count_max(asset), 1.0)
                )

                # Asset can be realized once the investments made equal the installation and
                # investment cost
                capex_sym = 0.0
                if self.variable_nominal(self._asset_installation_cost_map[asset]) > 1.0e2:
                    capex_sym = capex_sym + installation_cost_sym
                if self.variable_nominal(self._asset_investment_cost_map[asset]) > 1.0e2:
                    capex_sym = capex_sym + investment_cost_sym

                constraints.append(
                    (
                        (
                            cumulative_investments_made
                            - capex_sym
                            + (1.0 - asset_is_realized) * big_m
                        )
                        / nominal,
                        0.0,
                        np.inf,
                    )
                )

                # Once the asset is utilized the asset must be realized
                heat_flow = self.state(f"{asset}.Heat_flow")
                if not np.isinf(self.bounds()[f"{asset}.Heat_flow"][1]):
                    big_m = (
                        1.5
                        * self.bounds()[f"{asset}.Heat_flow"][1]
                        / max(self.get_aggregation_count_max(asset), 1.0)
                    )
                else:
                    try:
                        big_m = (
                            1.5
                            * max(
                                self.bounds()[f"{asset}.HeatOut.Heat"][1],
                                self.bounds()[f"{asset}.HeatIn.Heat"][1],
                            )
                            / max(self.get_aggregation_count_max(asset), 1.0)
                        )
                    except KeyError:
                        big_m = (
                            1.5
                            * max(
                                self.bounds()[f"{asset}.Primary.HeatOut.Heat"][1],
                                self.bounds()[f"{asset}.Primary.HeatIn.Heat"][1],
                            )
                            / max(self.get_aggregation_count_max(asset), 1.0)
                        )
                constraints.append(((heat_flow + asset_is_realized * big_m) / big_m, 0.0, np.inf))
                constraints.append(((heat_flow - asset_is_realized * big_m) / big_m, -np.inf, 0.0))

        return constraints

    def __annualized_capex_constraints(self, ensemble_member):
        """
        Calculate the annualized capital expenditure (CAPEX) constraints for different categories
        of assets in a milp network, taking into account the initial investment cost, the technical
        life of the asset, and the discount rate.
        The discount rate is used to calculate the periodic annual equivalent cost of the asset's
        capital investment over its technical life using the annuity formula.
        The discount rate reflects the time value of money and the risk associated with the
        investment. A higher discount rate will result in a lower present value of the annuity and
        a higher periodic payment, while a lower discount rate will result in a higher present
        value of the annuity and a lower periodic payment.

                Args:
                    ensemble_member: The ensemble member used to get parameters for the calculation.

                Returns:
                    A list of constraints for each asset.
        """
        constraints = []

        asset_categories = [
            "heat_source",
            "ates",
            "low_temperature_ates",
            "heat_buffer",
            "heat_pipe",
            "heat_exchanger",
            "heat_pump",
        ]

        parameters = super().parameters(ensemble_member)

        for category in asset_categories:
            for asset_name in self.energy_system_components.get(category, []):
                asset_life_years = parameters[f"{asset_name}.technical_life"]
                # Input is assumed as as annual percentage
                discount_percentage = parameters[f"{asset_name}.discount_rate"]
                if np.isnan(asset_life_years) or np.isnan(discount_percentage):
                    logger.warning(
                        f"Annualized cost cannot be computed for \
                              {asset_name} since technical_life \
                                or discount_rate are not set."
                    )
                    continue

                symbol_name = self._annualized_capex_var_map[asset_name]
                symbol = self.extra_variable(symbol_name)

                investment_cost_symbol_name = self._asset_investment_cost_map[asset_name]
                investment_cost_symbol = self.extra_variable(
                    investment_cost_symbol_name, ensemble_member
                )

                installation_cost_symbol_name = self._asset_installation_cost_map[asset_name]
                installation_cost_symbol = self.extra_variable(
                    installation_cost_symbol_name, ensemble_member
                )

                investment_and_installation_cost = investment_cost_symbol + installation_cost_symbol

                nominal = self.variable_nominal(symbol_name)
                discount_rate = discount_percentage / 100

                annuity_factor = calculate_annuity_factor(discount_rate, asset_life_years)

                constraints.append(
                    (
                        (symbol - investment_and_installation_cost * annuity_factor) / nominal,
                        0.0,
                        0.0,
                    )
                )

        return constraints

    def __revenue_constraints(self, ensemble_member):
        """
        TODO: Description revenue constraints
        """
        constraints = []
        # TODO: this workflow is still work in progress and this part of code still needs to be
        #  finalised

        # TODO: add fixed price default from ESDL in case no price profile is defined.
        parameters = self.parameters(ensemble_member)

        for demand in [
            *self.energy_system_components.get("gas_demand", []),
            *self.energy_system_components.get("electricity_demand", []),
        ]:

            carrier_name = None
            for _id, attr in self.get_electricity_carriers().items():
                if attr["id_number_mapping"] == parameters[f"{demand}.id_mapping_carrier"]:
                    carrier_name = attr["name"]
                    cost_multiplier = 1 / 3600.0  # priceprofile electricity is EUR/Wh
            for _id, attr in self.get_gas_carriers().items():
                if attr["id_number_mapping"] == parameters[f"{demand}.id_mapping_carrier"]:
                    carrier_name = attr["name"]
                    cost_multiplier = 1.0  # priceprofile gas is in EUR/g
            if carrier_name is not None:
                price_profile = self.get_timeseries(f"{carrier_name}.price_profile").values

                if demand in self.energy_system_components.get("gas_demand", []):
                    energy_flow = self.__state_vector_scaled(
                        f"{demand}.Gas_demand_mass_flow", ensemble_member  # g/s
                    )

                elif demand in self.energy_system_components.get("electricity_demand", []):
                    energy_flow = self.__state_vector_scaled(
                        f"{demand}.Electricity_demand", ensemble_member
                    )

                variable_revenue_var = self._asset_revenue_map[demand]
                variable_revenue = self.extra_variable(variable_revenue_var, ensemble_member)
                nominal = self.variable_nominal(variable_revenue_var)

                sum = 0.0
                timesteps = np.diff(self.times()) * cost_multiplier
                for i in range(1, len(self.times())):
                    sum += price_profile[i] * energy_flow[i] * timesteps[i - 1]

                constraints.append(((variable_revenue - sum) / (nominal), 0.0, 0.0))

        return constraints

    def path_constraints(self, ensemble_member):
        """
        Here we add all the path constraints to the optimization problem. Please note that the
        path constraints are the constraints that are applied to each time-step in the problem.
        """

        constraints = super().path_constraints(ensemble_member)

        constraints.extend(
            self.__cumulative_investments_made_in_eur_path_constraints(ensemble_member)
        )

        return constraints

    def constraints(self, ensemble_member):
        """
        This function adds the normal constraints to the problem. Unlike the path constraints these
        are not applied to every time-step in the problem. Meaning that these constraints either
        consider global variables that are independent of time-step or that the relevant time-steps
        are indexed within the constraint formulation.
        """
        constraints = super().constraints(ensemble_member)

        constraints.extend(self.__variable_operational_cost_constraints(ensemble_member))
        constraints.extend(self.__fixed_operational_cost_constraints(ensemble_member))
        constraints.extend(self.__investment_cost_constraints(ensemble_member))
        constraints.extend(self.__installation_cost_constraints(ensemble_member))
        constraints.extend(self.__revenue_constraints(ensemble_member))

        if self.energy_system_options()["discounted_annualized_cost"]:
            constraints.extend(self.__annualized_capex_constraints(ensemble_member))

        return constraints

    def goal_programming_options(self):
        """
        Here we set the goal programming configuration. We use soft constraints for consecutive
        goals.
        """
        options = super().goal_programming_options()
        options["keep_soft_constraints"] = True
        return options

    def solver_options(self):
        """
        Here we define the solver options. By default we use the open-source solver cbc and casadi
        solver qpsol.
        """
        options = super().solver_options()
        options["casadi_solver"] = "qpsol"
        options["solver"] = "highs"
        return options

    def compiler_options(self):
        """
        In this function we set the compiler configuration.
        """
        options = super().compiler_options()
        options["resolve_parameter_values"] = True
        return options


def calculate_annuity_factor(discount_rate: float, years_asset_life: float) -> float:
    """
    Calculate the annuity factor, given an annual discount_rate over a specified number
    of years_asset_life. This annuity factor is used in the model to calculate the
    annual constant payments equivalent to the initial investment or installation cost.
    Parameters:
        discount_rate (float): Annual discount rate (expressed
        as a decimal, e.g., 0.05 for 5%).
        years_asset_life (flor): Asset technical life (years).
    Returns:
        float: annuity_factor.
    """

    if discount_rate < 0 or discount_rate > 1:
        raise ValueError("Discount rate must be between 0-1")

    if years_asset_life <= 0:
        raise ValueError("Asset technical life must be greather than 0")

    if discount_rate == 0:
        annuity_factor = 1 / years_asset_life
    else:
        annuity_factor = discount_rate / (1 - (1 + discount_rate) ** (-years_asset_life))
    return annuity_factor



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile

import numpy as np

from rtctools.util import run_optimization_problem


class TestESDL(TestCase):
    def test_basic_source_and_demand_heat(self):
        """
        Check whether a hardcoded pycml model gives equivalent results compared to an esdl
        specified model. The model consists out of a source, pipe and demand

        Checks:
        - that the milp demand is equal.

        """
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatESDL, HeatPython

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent
        input_folder = base_folder / "input"

        case_python = run_optimization_problem(
            HeatPython, base_folder=base_folder, input_folder=input_folder
        )
        case_esdl = run_optimization_problem(
            HeatESDL,
            base_folder=base_folder,
            esdl_file_name="model.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.xml",
        )

        self.assertAlmostEqual(case_python.objective_value, case_esdl.objective_value, 6)

        np.testing.assert_allclose(
            case_python.extract_results()["demand.Heat_demand"],
            case_esdl.extract_results()["demand.Heat_demand"],
        )

    # def test_basic_source_and_demand_qth(self):
    #     import models.basic_source_and_demand.src.qth_comparison as qth_comparison
    #     from models.basic_source_and_demand.src.qth_comparison import QTHPython, QTHESDL
    #
    #     base_folder = Path(qth_comparison.__file__).resolve().parent.parent
    #
    #     case_python = run_optimization_problem(QTHPython, base_folder=base_folder)
    #     case_esdl = run_optimization_problem(QTHESDL, base_folder=base_folder)
    #
    #     np.testing.assert_allclose(
    #         case_python._objective_values[0], case_esdl._objective_values, rtol=1e-5, atol=1e-5
    #     )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestESDL()
    a.test_basic_source_and_demand_heat()
    temp = 0



import logging

import casadi as ca

from mesido.base_component_type_mixin import BaseComponentTypeMixin
from mesido.head_loss_class import HeadLossClass, HeadLossOption
from mesido.network_common import NetworkSettings

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.timeseries import Timeseries

logger = logging.getLogger("mesido")


class GasPhysicsMixin(BaseComponentTypeMixin, CollocatedIntegratedOptimizationProblem):
    __allowed_head_loss_options = {
        HeadLossOption.NO_HEADLOSS,
        HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY,
        HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY,
    }
    """
    This class is used to model the physics of a gas network with its assets. We model
    the different components with variety of linearization strategies.
    """

    def __init__(self, *args, **kwargs):
        r"""
        In this __init__ we prepare the dicts for the variables added by the HeatMixin class

        Gas network specific settings:

        The ``network_type`` is the network type identifier.

        The ``maximum_velocity`` is the maximum absolute value of the velocity in every pipe. This
        velocity is also used in the head loss / hydraulic power to calculate the maximum discharge
        if no maximum velocity per pipe class is not specified.

        The ``minimum_velocity`` is the minimum absolute value of the velocity
        in every pipe. It is mostly an option to improve the stability of the
        solver in a possibly subsequent QTH problem: the default value of
        `0.005` m/s helps the solver by avoiding the difficult case where
        discharges get close to zero.

        To model the head loss in pipes, the ``head_loss_option`` refers to
        one of the ways this can be done. See :class:`HeadLossOption` for more
        explanation on what each option entails. Note that some options model
        the head loss as an inequality, i.e. :math:`\Delta H \ge f(Q)`, whereas
        others model it as an equality.

        When ``HeadLossOption.CQ2_INEQUALITY`` is used, the wall roughness at
        ``estimated_velocity`` determines the `C` in :math:`\Delta H \ge C
        \cdot Q^2`.

        When ``HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY`` is used, the
        ``maximum_velocity`` needs to be set. The Darcy-Weisbach head loss
        relationship from :math:`v = 0` until :math:`v = \text{maximum_velocity}`
        will then be linearized using ``n_linearization`` lines.

        When ``HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY`` is used, the wall roughness at
        ``estimated_velocity`` determines the `C` in :math:`\Delta H = C \cdot
        Q`. For pipes that contain a control valve, the formulation of
        ``HeadLossOption.CQ2_INEQUALITY`` is used.

        When ``HeadLossOption.CQ2_EQUALITY`` is used, the wall roughness at
        ``estimated_velocity`` determines the `C` in :math:`\Delta H = C \cdot
        Q^2`. Note that this formulation is non-convex. At `theta < 1` we
        therefore use the formulation ``HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY``. For pipes
        that contain a control valve, the formulation of
        ``HeadLossOption.CQ2_INEQUALITY`` is used.

        When ``minimize_head_losses`` is set to True (default), a last
        priority is inserted where the head losses and hydraulic power in the system are
        minimized if the ``head_loss_option`` is not `NO_HEADLOSS`.
        This is related to the assumption that control valves are
        present in the system to steer water in the right direction the case
        of multiple routes. If such control valves are not present, enabling
        this option will give warnings in case the found solution is not
        feasible. In case the option is False, both the minimization and
        checks are skipped.

        Note that the inherited options ``head_loss_option`` and
        ``minimize_head_losses`` are changed from their default values to
        ``HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY`` and ``False`` respectively.

        The ``n_linearization_lines`` is the number of lines used when a curve is approximated by
        multiple linear lines.

        The ``pipe_minimum_pressure`` is the global minimum pressured allowed
        in the network. Similarly, ``pipe_maximum_pressure`` is the maximum
        one.
        """
        self.gas_network_settings = {
            "network_type": NetworkSettings.NETWORK_TYPE_GAS,
            "maximum_velocity": 15.0,
            "minimum_velocity": 0.005,
            "head_loss_option": HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY,
            "minimize_head_losses": False,
            "n_linearization_lines": 5,
            "pipe_minimum_pressure": -np.inf,
            "pipe_maximum_pressure": np.inf,
        }
        self._gn_head_loss_class = HeadLossClass(self.gas_network_settings)
        self.__gas_pipe_head_bounds = {}
        self.__gas_pipe_head_loss_var = {}
        self.__gas_pipe_head_loss_bounds = {}
        self.__gas_pipe_head_loss_nominals = {}
        self.__gas_pipe_head_loss_zero_bounds = {}
        self._gn_pipe_to_head_loss_map = {}

        # Boolean path-variable for the direction of the flow, inport to outport is positive flow.
        self.__gas_flow_direct_var = {}
        self.__gas_flow_direct_bounds = {}
        self._gas_pipe_to_flow_direct_map = {}

        # Still to be implemented
        # Boolean path-variable to determine whether flow is going through a pipe.
        # self.__gas_pipe_disconnect_var = {}
        # self.__gas_pipe_disconnect_var_bounds = {}
        # self._gas_pipe_disconnect_map = {}

        # Boolean variables for the linear line segment options per pipe.
        # TDOD: change name to _gas_pipe_...
        self.__gas_pipe_linear_line_segment_var = {}  # value 0/1: line segment - not active/active
        self.__gas_pipe_linear_line_segment_var_bounds = {}
        self._gas_pipe_linear_line_segment_map = {}

        super().__init__(*args, **kwargs)

        self._gas_pipe_topo_pipe_class_map = {}

        # self.__gas_pipe_disconnect_var = {}
        # self.__gas_pipe_disconnect_var_bounds = {}
        self._gas_pipe_disconnect_map = {}

        self.__gas_storage_discharge_var = {}
        self.__gas_storage_discharge_bounds = {}
        self.__gas_storage_discharge_nominals = {}
        self.__gas_storage_discharge_map = {}

    def gas_carriers(self):
        """
        This function should be overwritten by the problem and should give a dict with the
        carriers as keys and a list of temperatures as values.
        """
        return {}

    def pre(self):
        """
        In this pre method we fill the dicts initiated in the __init__. This means that we create
        the Casadi variables and determine the bounds, nominals and create maps for easier
        retrieving of the variables.
        """
        super().pre()

        options = self.energy_system_options()

        def _get_max_bound(bound):
            if isinstance(bound, np.ndarray):
                return max(bound)
            elif isinstance(bound, Timeseries):
                return max(bound.values)
            else:
                return bound

        def _get_min_bound(bound):
            if isinstance(bound, np.ndarray):
                return min(bound)
            elif isinstance(bound, Timeseries):
                return min(bound.values)
            else:
                return bound

        bounds = self.bounds()

        for pipe_name in self.energy_system_components.get("gas_pipe", []):
            head_loss_var = f"{pipe_name}.__head_loss"
            # Note we always use the gas network type for the naming of variables, independent of
            # the gas mixture used.
            initialized_vars = self._gn_head_loss_class.initialize_variables_nominals_and_bounds(
                self,
                self.gas_network_settings["network_type"],
                pipe_name,
                self.gas_network_settings,
            )
            if initialized_vars[0] != {}:
                self.__gas_pipe_head_bounds[
                    f"{pipe_name}.{NetworkSettings.NETWORK_TYPE_GAS}In.H"
                ] = initialized_vars[0]
            if initialized_vars[1] != {}:
                self.__gas_pipe_head_bounds[
                    f"{pipe_name}.{NetworkSettings.NETWORK_TYPE_GAS}Out.H"
                ] = initialized_vars[1]
            if initialized_vars[2] != {}:
                self.__gas_pipe_head_loss_zero_bounds[f"{pipe_name}.dH"] = initialized_vars[2]
            if initialized_vars[3] != {}:
                self._gn_pipe_to_head_loss_map[pipe_name] = initialized_vars[3]
            if initialized_vars[4] != {}:
                self.__gas_pipe_head_loss_var[head_loss_var] = initialized_vars[4]
            if initialized_vars[5] != {}:
                self.__gas_pipe_head_loss_nominals[f"{pipe_name}.dH"] = initialized_vars[5]
            if initialized_vars[6] != {}:
                self.__gas_pipe_head_loss_nominals[head_loss_var] = initialized_vars[6]
            if initialized_vars[7] != {}:
                self.__gas_pipe_head_loss_bounds[head_loss_var] = initialized_vars[7]

            if (
                initialized_vars[8] != {}
                and initialized_vars[9] != {}
                and initialized_vars[10] != {}
            ):  # Variables needed to indicate if a linear line segment is active
                self._gas_pipe_linear_line_segment_map[pipe_name] = {}
                for ii_line in range(self.gas_network_settings["n_linearization_lines"] * 2):
                    pipe_linear_line_segment_var_name = initialized_vars[8][ii_line]
                    self._gas_pipe_linear_line_segment_map[pipe_name][
                        ii_line
                    ] = pipe_linear_line_segment_var_name
                    self.__gas_pipe_linear_line_segment_var[pipe_linear_line_segment_var_name] = (
                        initialized_vars[9][pipe_linear_line_segment_var_name]
                    )
                    self.__gas_pipe_linear_line_segment_var_bounds[
                        pipe_linear_line_segment_var_name
                    ] = initialized_vars[10][pipe_linear_line_segment_var_name]

            # Integer variables
            flow_dir_var = f"{pipe_name}__gas_flow_direct_var"

            self._gas_pipe_to_flow_direct_map[pipe_name] = flow_dir_var
            self.__gas_flow_direct_var[flow_dir_var] = ca.MX.sym(flow_dir_var)

            # Fix the directions that are already implied by the bounds on milp
            # Nonnegative milp implies that flow direction Boolean is equal to one.
            # Nonpositive milp implies that flow direction Boolean is equal to zero.

            q_in_lb = _get_min_bound(bounds[f"{pipe_name}.GasIn.Q"][0])
            q_in_ub = _get_max_bound(bounds[f"{pipe_name}.GasIn.Q"][1])
            q_out_lb = _get_min_bound(bounds[f"{pipe_name}.GasOut.Q"][0])
            q_out_ub = _get_max_bound(bounds[f"{pipe_name}.GasOut.Q"][1])

            if (q_in_lb >= 0.0 and q_in_ub >= 0.0) or (q_out_lb >= 0.0 and q_out_ub >= 0.0):
                self.__gas_flow_direct_bounds[flow_dir_var] = (1.0, 1.0)
            elif (q_in_lb <= 0.0 and q_in_ub <= 0.0) or (q_out_lb <= 0.0 and q_out_ub <= 0.0):
                self.__gas_flow_direct_bounds[flow_dir_var] = (0.0, 0.0)
            else:
                self.__gas_flow_direct_bounds[flow_dir_var] = (0.0, 1.0)

            # # Still to be added in the future
            # if parameters[f"{pipe_name}.disconnectable"]:
            #     disconnected_var = f"{pipe_name}__is_disconnected"
            #     self._gas_pipe_disconnect_map[pipe_name] = disconnected_var
            #     self.__gas_pipe_disconnect_var[disconnected_var] = ca.MX.sym(disconnected_var)
            #     self.__gas_pipe_disconnect_var_bounds[disconnected_var] = (0.0, 1.0)

        self.__maximum_total_head_loss = self.__get_maximum_total_head_loss()

        if options["gas_storage_discharge_variables"]:
            for storage in self.energy_system_components.get("gas_tank_storage", []):
                bound_storage_q = self.bounds()[f"{storage}.GasIn.Q"][0]
                var_name = f"{storage}__Q_discharge"
                self.__gas_storage_discharge_map[storage] = var_name
                self.__gas_storage_discharge_var[var_name] = ca.MX.sym(var_name)
                self.__gas_storage_discharge_bounds[var_name] = (0, -bound_storage_q)
                self.__gas_storage_discharge_nominals[var_name] = self.variable_nominal(
                    f"{storage}.GasIn.Q"
                )

    def energy_system_options(self):
        r"""
        Returns a dictionary of milp network specific options.

        gas_storage_discharge_variables: creates separate variables for the discharge of the
        gas_storages, only required when using the multicommodity simulator as this requires
        separate variables to create goals with.
        """

        options = self._gn_head_loss_class.head_loss_network_options()

        options["minimum_pressure_far_point"] = 1.0
        options["gas_storage_discharge_variables"] = False

        return options

    @property
    def extra_variables(self):
        """
        In this function we add all the variables defined in the HeatMixin to the optimization
        problem. Note that these are only the normal variables not path variables.
        """
        variables = super().extra_variables.copy()

        return variables

    @property
    def path_variables(self):
        """
        In this function we add all the path variables defined in the HeatMixin to the
        optimization problem. Note that path_variables are variables that are created for each
        time-step.
        """
        variables = super().path_variables.copy()
        variables.extend(self.__gas_pipe_head_loss_var.values())
        variables.extend(self.__gas_flow_direct_var.values())
        # variables.extend(self.__gas_pipe_disconnect_var.values())  # still to be implemented
        variables.extend(self.__gas_pipe_linear_line_segment_var.values())
        variables.extend(self.__gas_storage_discharge_var.values())

        return variables

    def variable_is_discrete(self, variable):
        """
        All variables that only can take integer values should be added to this function.
        """
        if (
            variable in self.__gas_flow_direct_var
            or variable in self.__gas_pipe_linear_line_segment_var
        ):
            return True
        else:
            return super().variable_is_discrete(variable)

    def variable_nominal(self, variable):
        """
        In this function we add all the nominals for the variables defined/added in the HeatMixin.
        """

        if variable in self.__gas_pipe_head_loss_nominals:
            return self.__gas_pipe_head_loss_nominals[variable]
        elif variable in self.__gas_storage_discharge_nominals:
            return self.__gas_storage_discharge_nominals[variable]
        else:
            return super().variable_nominal(variable)

    def bounds(self):
        """
        In this function we add the bounds to the problem for all the variables defined/added in
        the HeatMixin.
        """
        bounds = super().bounds()

        bounds.update(self.__gas_flow_direct_bounds)
        # bounds.update(self.__gas_pipe_disconnect_var_bounds)  # still to be implemented
        bounds.update(self.__gas_pipe_head_loss_bounds)
        bounds.update(self.__gas_pipe_head_loss_zero_bounds)
        bounds.update(self.__gas_pipe_linear_line_segment_var_bounds)
        bounds.update(self.__gas_storage_discharge_bounds)

        for k, v in self.__gas_pipe_head_bounds.items():
            bounds[k] = self.merge_bounds(bounds[k], v)

        return bounds

    def path_goals(self):
        """
        Here we add the goals for minimizing the head loss and hydraulic power depending on the
        configuration. Please note that we only do hydraulic power for the MILP problem thus only
        for the linearized head_loss options.
        """
        g = super().path_goals().copy()

        if (
            self.gas_network_settings["minimize_head_losses"]
            and self.gas_network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS
        ):
            g.append(
                self._gn_head_loss_class._hn_minimization_goal_class(
                    self, self.gas_network_settings
                )
            )
            if (
                self.gas_network_settings["head_loss_option"]
                == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY
                or self.gas_network_settings["head_loss_option"]
                == HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
            ):
                g.append(
                    self._gn_head_loss_class._hpwr_minimization_goal_class(
                        self,
                        self.gas_network_settings,
                    )
                )

        return g

    def __get_maximum_total_head_loss(self):
        """
        Get an upper bound on the maximum total head loss that can be used in
        big-M formulations of e.g. check valves and disconnectable pipes.

        There are multiple ways to calculate this upper bound, depending on
        what options are set. We compute all these upper bounds, and return
        the lowest one of them.
        """

        options = self.energy_system_options()
        components = self.energy_system_components

        if self.gas_network_settings["head_loss_option"] == HeadLossOption.NO_HEADLOSS:
            # Undefined, and all constraints using this methods value should
            # be skipped.
            return np.nan

        # Summing head loss in pipes
        max_sum_dh_pipes = 0.0

        for ensemble_member in range(self.ensemble_size):
            parameters = self.parameters(ensemble_member)

            head_loss = 0.0

            for pipe in components.get("gas_pipe", []):
                area = parameters[f"{pipe}.area"]
                max_discharge = self.gas_network_settings["maximum_velocity"] * area
                head_loss += self._gn_head_loss_class._hn_pipe_head_loss(
                    pipe,
                    self,
                    options,
                    self.gas_network_settings,
                    parameters,
                    max_discharge,
                    # network_type=self.gas_network_settings["network_type"],
                    pressure=parameters[f"{pipe}.pressure"],
                )

            head_loss += options["minimum_pressure_far_point"] * 10.2

            max_sum_dh_pipes = max(max_sum_dh_pipes, head_loss)

        # Maximum pressure difference allowed with user options
        # NOTE: Does not yet take elevation differences into acccount
        max_dh_network_options = (
            self.gas_network_settings["pipe_maximum_pressure"]
            - self.gas_network_settings["pipe_minimum_pressure"]
        ) * 10.2

        return min(max_sum_dh_pipes, max_dh_network_options)

    def __gas_node_mixing_path_constraints(self, ensemble_member):
        """
        This function adds constraints for each gas network node/joint to have as much
        flow going in as out. Effectively, it is setting the sum of flow to zero.
        """
        constraints = []

        for node, connected_pipes in self.energy_system_topology.gas_nodes.items():
            q_sum = 0.0
            q_nominals = []

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                gas_conn = f"{node}.GasConn[{i_conn + 1}].Q"
                q_sum += orientation * self.state(gas_conn)
                q_nominals.append(self.variable_nominal(gas_conn))

            q_nominal = np.median(q_nominals)
            constraints.append((q_sum / q_nominal, 0.0, 0.0))

            q_sum = 0.0
            q_nominals = []

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                gas_conn = f"{node}.GasConn[{i_conn + 1}].Q_shadow"
                q_sum += orientation * self.state(gas_conn)
                q_nominals.append(self.variable_nominal(gas_conn))

            q_nominal = np.median(q_nominals)
            constraints.append((q_sum / q_nominal, 0.0, 0.0))

        return constraints

    def __gas_node_hydraulic_power_mixing_path_constraints(self, ensemble_member):
        """
        This function adds constraints to ensure that the incoming hydraulic power equals the
        outgoing hydraulic power. We assume constant density throughout a hydraulically coupled
        system and thus these constraints are needed for mass conservation.
        """
        constraints = []

        for node, connected_pipes in self.energy_system_topology.gas_nodes.items():
            q_sum = 0.0
            q_nominals = []

            for i_conn, (_pipe, orientation) in connected_pipes.items():
                q_conn = f"{node}.GasConn[{i_conn + 1}].Hydraulic_power"
                q_sum += orientation * self.state(q_conn)
                q_nominals.append(self.variable_nominal(q_conn))

            q_nominal = np.median(q_nominals)
            constraints.append((q_sum / q_nominal, 0.0, 0.0))

        return constraints

    def __state_vector_scaled(self, variable, ensemble_member):
        """
        This functions returns the casadi symbols scaled with their nominal for the entire time
        horizon.
        """
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def __flow_direction_path_constraints(self, ensemble_member):
        """
        This function adds constraints to set the direction in pipes and determine whether a pipe
        is utilized at all (is_disconnected variable).

        Whether a pipe is connected is based upon whether flow passes through that pipe.

        The directions are set based upon the directions of how thermal power propegates. This is
        done based upon the sign of the Heat variable. Where positive Heat means a positive
        direction and negative milp means a negative direction. By default, positive is defined from
        HeatIn to HeatOut.

        Finally, a minimum flow can be set. This can sometimes be useful for numerical stability.
        """
        constraints = []

        # Also ensure that the discharge has the same sign as the milp.
        for p in self.energy_system_components.get("gas_pipe", []):
            flow_dir_var = self._gas_pipe_to_flow_direct_map[p]
            flow_dir = self.state(flow_dir_var)

            q_in = self.state(f"{p}.GasIn.Q")

            big_m = 2.0 * np.max(
                np.abs(
                    (
                        *self.bounds()[f"{p}.GasIn.Q"],
                        *self.bounds()[f"{p}.GasOut.Q"],
                    )
                )
            )

            # Note we only need one on the milp as the desired behaviour is propegated by the
            # constraints heat_in - heat_out - heat_loss == 0.
            constraints.append(
                (
                    (q_in - big_m * flow_dir) / big_m,
                    -np.inf,
                    0.0,
                )
            )
            constraints.append(
                (
                    (q_in + big_m * (1 - flow_dir)) / big_m,
                    0.0,
                    np.inf,
                )
            )

        # Pipes that are connected in series should have the same milp direction.
        for pipes in self.energy_system_topology.pipe_series:
            if len(pipes) <= 1:
                continue

            base_flow_dir_var = self.state(self._gas_pipe_to_flow_direct_map[pipes[0]])

            for p in pipes[1:]:
                flow_dir_var = self.state(self._gas_pipe_to_flow_direct_map[p])
                constraints.append((base_flow_dir_var - flow_dir_var, 0.0, 0.0))

        return constraints

    def __gas_storage_discharge_path_constraints(self, ensemble_member):
        """
        The discharge variables are added such that two separate goals for charging and discharging
        can be created. The discharging variable has a lower bound of 0 and should always be larger
        or equal to the negative of the inflow variable. This allows for first a minimization of
        the discharging and afterwards an maximisation of the charging without conflicting goals or
        constraints.
        """
        constraints = []
        options = self.energy_system_options()

        if options["gas_storage_discharge_variables"]:
            for storage in self.energy_system_components.get("gas_tank_storage", []):
                storage_charge_var = self.state(f"{storage}.GasIn.Q")
                storage_discharge_var_name = f"{storage}__Q_discharge"
                storage_discharge_var = self.state(storage_discharge_var_name)
                nominal = self.variable_nominal(storage_discharge_var_name)

                # Q_discharge >= -Q
                constraints.append(
                    ((storage_discharge_var + storage_charge_var) / nominal, 0.0, np.inf)
                )

        return constraints

    def path_constraints(self, ensemble_member):
        """
        Here we add all the path constraints to the optimization problem. Please note that the
        path constraints are the constraints that are applied to each time-step in the problem.
        """

        constraints = super().path_constraints(ensemble_member)

        constraints.extend(self.__gas_node_mixing_path_constraints(ensemble_member))

        # Add source/demand head loss constrains only if head loss is non-zero
        if self.gas_network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS:
            constraints.extend(
                self._gn_head_loss_class._pipe_head_loss_path_constraints(self, ensemble_member)
            )
        constraints.extend(
            self._gn_head_loss_class._pipe_hydraulic_power_path_constraints(
                self, self.__maximum_total_head_loss, ensemble_member
            )
        )
        constraints.extend(self.__flow_direction_path_constraints(ensemble_member))
        constraints.extend(self.__gas_node_hydraulic_power_mixing_path_constraints(ensemble_member))
        constraints.extend(self.__gas_storage_discharge_path_constraints(ensemble_member))

        return constraints

    def constraints(self, ensemble_member):
        """
        This function adds the normal constraints to the problem. Unlike the path constraints these
        are not applied to every time-step in the problem. Meaning that these constraints either
        consider global variables that are independent of time-step or that the relevant time-steps
        are indexed within the constraint formulation.
        """
        constraints = super().constraints(ensemble_member)

        if self.gas_network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS:
            constraints.extend(
                self._gn_head_loss_class._pipe_head_loss_constraints(
                    self, self.__maximum_total_head_loss, ensemble_member
                )
            )

        return constraints

    def goal_programming_options(self):
        """
        Here we set the goal programming configuration. We use soft constraints for consecutive
        goals.
        """
        options = super().goal_programming_options()
        options["keep_soft_constraints"] = True
        return options

    def solver_options(self):
        """
        Here we define the solver options. By default we use the open-source solver cbc and casadi
        solver qpsol.
        """
        options = super().solver_options()
        options["casadi_solver"] = "qpsol"
        options["solver"] = "highs"
        return options

    def compiler_options(self):
        """
        In this function we set the compiler configuration.
        """
        options = super().compiler_options()
        options["resolve_parameter_values"] = True
        return options

    def priority_completed(self, priority):
        """
        This function is called after a priority of goals is completed. This function is used to
        specify operations between consecutive goals. Here we set some parameter attributes after
        the optimization is completed.
        """
        options = self.energy_system_options()

        if (
            self.gas_network_settings["minimize_head_losses"]
            and self.gas_network_settings["head_loss_option"] != HeadLossOption.NO_HEADLOSS
            and priority == self._gn_head_loss_class._hn_minimization_goal_class.priority
        ):
            components = self.energy_system_components

            rtol = 1e-5
            atol = 1e-4

            for ensemble_member in range(self.ensemble_size):
                parameters = self.parameters(ensemble_member)
                results = self.extract_results(ensemble_member)

                for pipe in components.get("gas_pipe", []):
                    # if parameters[f"{pipe}.has_control_valve"]: # not used at all
                    #     continue

                    # Just like with a control valve, if pipe is disconnected
                    # there is nothing to check.
                    q_full = results[f"{pipe}.Q"]
                    # if parameters[f"{pipe}.disconnectable"]: # not used yet
                    #     inds = q_full != 0.0
                    # else:
                    #     inds = np.arange(len(q_full), dtype=int)
                    inds = np.arange(len(q_full), dtype=int)

                    if parameters[f"{pipe}.diameter"] == 0.0:
                        # Pipe is disconnected. Head loss is free, so nothing to check.
                        continue

                    q = results[f"{pipe}.Q"][inds]
                    head_loss_target = self._gn_head_loss_class._hn_pipe_head_loss(
                        pipe,
                        self,
                        options,
                        self.gas_network_settings,
                        parameters,
                        q,
                        None,
                        # network_type=self.gas_network_settings["network_type"],
                        pressure=parameters[f"{pipe}.pressure"],
                    )
                    if (
                        self.gas_network_settings["head_loss_option"]
                        == HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY
                    ):
                        head_loss = np.abs(results[f"{pipe}.dH"][inds])
                    else:
                        head_loss = results[self._gn_pipe_to_head_loss_map[pipe]][inds]

                    if not np.allclose(head_loss, head_loss_target, rtol=rtol, atol=atol):
                        logger.warning(
                            f"Pipe {pipe} has artificial head loss; "
                            f"at least one more control valve should be added to the network."
                        )

        super().priority_completed(priority)



from .grow_workflow import (
    EndScenarioSizing,
    EndScenarioSizingDiscountedHIGHS,
    EndScenarioSizingDiscountedStagedHIGHS,
    EndScenarioSizingHIGHS,
    EndScenarioSizingStaged,
    EndScenarioSizingStagedHIGHS,
    SolverGurobi,
    SolverHIGHS,
    run_end_scenario_sizing,
    run_end_scenario_sizing_no_heat_losses,
)
from .simulator_workflow import (
    NetworkSimulator,
    NetworkSimulatorHIGHS,
    NetworkSimulatorHIGHSTestCase,
    NetworkSimulatorHIGHSWeeklyTimeStep,
)


__all__ = [
    "EndScenarioSizing",
    "EndScenarioSizingDiscountedHIGHS",
    "EndScenarioSizingDiscountedStagedHIGHS",
    "EndScenarioSizingHIGHS",
    "EndScenarioSizingStaged",
    "EndScenarioSizingStagedHIGHS",
    "SolverGurobi",
    "SolverHIGHS",
    "run_end_scenario_sizing",
    "run_end_scenario_sizing_no_heat_losses",
    "NetworkSimulator",
    "NetworkSimulatorHIGHS",
    "NetworkSimulatorHIGHSTestCase",
    "NetworkSimulatorHIGHSWeeklyTimeStep",
]



from mesido.pycml.component_library.milp._internal.heat_component import BaseAsset


class ElectricityComponent(BaseAsset):
    """
    Base electricity component nothing to add here yet.
    """

    pass



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

import pytest


class TestSetpointConstraints(TestCase):
    @pytest.mark.first
    def test_setpoint_constraints(self):
        """
        his function checks the working of the setpoint constraints for a few cases to ensure the
        behaviour is as expected. The setpoint constraints are used to enforce a maximum number of
        changes allowed of X times over a desired window length.

        Checks:
        - That setpoint does not change over windowlength if 0 is specified.
        - That setpoint does not change over windowlength if 0 is specified via the
        ESDLAdditionalVarsMixin.
        - That setpoint indeed changes once if 1 is specified.

        """
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemSetPointConstraints

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        _heat_problem_3 = run_esdl_mesido_optimization(
            HeatProblemSetPointConstraints,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
            **{"timed_setpoints": {"GeothermalSource_b702": (45, 1)}},
        )
        results_3 = _heat_problem_3.extract_results()

        _heat_problem_4 = run_esdl_mesido_optimization(
            HeatProblemSetPointConstraints,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
            **{"timed_setpoints": {"GeothermalSource_b702": (45, 0)}},
        )
        results_4 = _heat_problem_4.extract_results()

        # Here we check whehter the ESDLAdditionalVarsMixin is working as intended when
        # a constraint for the setpoints is specified
        import models.unit_cases.case_3a_setpoint.src.run_3a as run_3a
        from models.unit_cases.case_3a_setpoint.src.run_3a import HeatProblem

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        sol_esdl_setpoints = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        results_4 = _heat_problem_4.extract_results()

        esdl_results = sol_esdl_setpoints.extract_results()
        np.testing.assert_array_less(
            abs(
                esdl_results["GeothermalSource_b702.Heat_source"][2:]
                - esdl_results["GeothermalSource_b702.Heat_source"][1:-1]
            ),
            1.0e-6,
        )

        # Check that solution has one setpoint change
        a = abs(
            results_3["GeothermalSource_b702.Heat_source"][2:]
            - results_3["GeothermalSource_b702.Heat_source"][1:-1]
        )
        np.testing.assert_array_less((a >= 1.0).sum(), 2)  # the 1.0 value is a manual threshold

        # Check that solution has no setpoint change
        np.testing.assert_array_less(
            abs(
                results_4["GeothermalSource_b702.Heat_source"][2:]
                - results_4["GeothermalSource_b702.Heat_source"][1:-1]
            ),
            1.0e-3,
        )

    @pytest.mark.second
    def test_run_small_ates_timed_setpoints_2_changes(self):
        """
        Run the small network with ATES and check that the setpoint changes as specified.
        The heat source for producer_1 changes 8 times (consecutively) when no timed_setpoints are
        specified. The 1 year heat demand profiles contains demand values: hourly (peak day), weekly
        (every 5days/120hours/432000s) and 1 time step of 4days (96hours/345600s, step before the
        start of the peak day). Now check that the time_setpoints can limit the setpoint changes to
        2 changes/year.

        Checks:
        - That the setpoint does change twice over window length if 2 is specified

        """
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import HeatProblemSetPoints

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblemSetPoints,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            **{"timed_setpoints": {"HeatProducer_1": (24 * 365, 2)}},
        )
        results = solution.extract_results()
        check = abs(
            results["HeatProducer_1.Heat_source"][2:] - results["HeatProducer_1.Heat_source"][1:-1]
        )
        # check if there are less than 3 switches, a solution might be found with less
        # than 2 switches
        np.testing.assert_array_less((check >= 1.0).sum(), 3.0)

    @pytest.mark.third
    def test_run_small_ates_timed_setpoints_0_changes(self):
        """
        Run the small network with ATES and check that the setpoint changes as specified.
        The heat source for producer_1 changes 8 times (consecutively) when no timed_setpoints are
        specified. The 1 year heat demand profiles contains demand values: hourly (peak day), weekly
        (every 5days/120hours/432000s) and 1 time step of 4days (96hours/345600s, step before the
        start of the peak day). Now check that the time_setpoints can limit the setpoint changes to
        0 changes/year.

        Checks:
        - That setpoint does not change over window length if 0 is specified

        """
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import HeatProblemSetPoints

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblemSetPoints,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            **{"timed_setpoints": {"HeatProducer_1": (24 * 365, 0)}},  # not change at all - works
        )
        results = solution.extract_results()
        check = abs(
            results["HeatProducer_1.Heat_source"][2:] - results["HeatProducer_1.Heat_source"][1:-1]
        )
        np.testing.assert_array_less(check, 1.0e-5)

    @pytest.mark.fourth
    def test_run_small_ates_timed_setpoints_multiple_constraints(self):
        """
        Run the small network with ATES and check that the setpoint changes as specified.
        The heat source for producer_1 changes 8 times (consecutively) when no timed_setpoints are
        specified. The 1 year heat demand profiles contains demand values: hourly (peak day), weekly
        (every 5days/120hours/432000s) and 1 time step of 4days (96hours/345600s, step before the
        start of the peak day). Now check that the time_setpoints can limit the setpoint changes to
        1 changes over multiple window sizes.

        Checks:
        - That setpoint does change once over window length if 1 is specified for multiple
        window sizes

        """
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import HeatProblemSetPoints

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        for ihrs in range(119, 122):
            solution = run_esdl_mesido_optimization(
                HeatProblemSetPoints,
                base_folder=base_folder,
                esdl_file_name="test_case_small_network_with_ates.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="Warmte_test.csv",
                **{"timed_setpoints": {"HeatProducer_1": (ihrs, 1)}},
            )
            results = solution.extract_results()
            diff = (
                results["HeatProducer_1.Heat_source"][2:]
                - results["HeatProducer_1.Heat_source"][1:-1]
            )
            ires = [idx + 2 for idx, val in enumerate(diff) if abs(val) > 1e-6]
            for ii in range(1, len(ires)):
                check = (
                    solution.get_timeseries("HeatingDemand_1.target_heat_demand", 0).times[ires[ii]]
                    - solution.get_timeseries("HeatingDemand_1.target_heat_demand", 0).times[
                        ires[ii - 1]
                    ]
                )
                # The following checks should be true because the changes setpoint changes occur at
                # the during 120hr time intervals
                if ihrs == 119:
                    np.testing.assert_array_less(ihrs * 3600, check)
                elif ihrs == 120:
                    np.testing.assert_equal(np.less_equal(ihrs * 3600, check), True)
                elif ihrs == 121:
                    np.testing.assert_equal(np.less_equal((ihrs - 1) * 3600, check), True)
                else:
                    exit("ii out of range")


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestSetpointConstraints()
    a.test_setpoint_constraints()
    a.test_run_small_ates_timed_setpoints_2_changes()
    a.test_run_small_ates_timed_setpoints_0_changes()
    a.test_run_small_ates_timed_setpoints_multiple_constraints()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))
